#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Exports the evaluate function, which takes a list of edit events
as well as the original dataframe, and returns the current state 
of the sheet as a dataframe
"""
import re
from itertools import chain

from mitosheet.errors import make_invalid_formula_error

def get_string_ranges(formula):
    """
    Returns a list of all the ranges that are strings within a formula.

    Useful for other functions that want to "not change strings."
    """
    if formula is None:
        return None

    # We find the ranges of the formula that are string constants;
    # we do not edit these! Taken from: https://stackoverflow.com/a/63707053/13113837
    string_matches_double_quotes = re.finditer(r'"(?:(?:(?!(?<!\\)").)*)"', formula)
    string_matches_single_quotes = re.finditer(r'\'(?:(?:(?!(?<!\\)\').)*)\'', formula)
    string_ranges = []
    for string_match in chain(string_matches_double_quotes, string_matches_single_quotes):
        string_ranges.append(string_match.span())

    return string_ranges

def match_in_strings(string_ranges, match):
    """
    Returns True iff a given match is within one of the string ranges.
    """
    text: str = match.group()
    start = match.start()
    end = match.end() # this is +1 after the last char of the string

    # We check if it's in any of the string ranges
    for string_range in string_ranges:
        string_start = string_range[0]
        string_end = string_range[1]

        if (string_start <= start and string_end >= end):
            return True

    return False

def safe_contains(formula, substring):
    """
    Returns true if the formula contains substring. However, will not count
    the substring if it is inside of a string.

    NOTE: this assumes the formula is valid!
    """

    # we do not search inside the string literals! 
    string_ranges = get_string_ranges(formula)

    for match in re.finditer(substring, formula):
        if not match_in_strings(string_ranges, match):
            return True

    return False

def safe_contains_function(formula, function):
    """
    Checks if a function is called in a formula.
    """
    formula = formula.upper()
    function = function.upper()

    # we do not search inside the string literals! 
    string_ranges = get_string_ranges(formula)

    for match in re.finditer(function, formula):
        if not match_in_strings(string_ranges, match):
            # Check if this is a function
            text: str = match.group()
            end = match.end() # this is +1 after the last char of the string
            if end < len(formula) and formula[end] == '(':
                return True

    return False


def safe_count_function(formula, substring):
    """
    Counts the number of occurences of the substring in the formula
    not including the string ranges

    NOTE: is case insensitive!
    """
    formula = formula.upper()
    function = substring.upper()

    # we do not search inside the string literals! 
    string_ranges = get_string_ranges(formula)

    count = 0
    for match in re.finditer(function, formula):
        if not match_in_strings(string_ranges, match):
            count += 1
    
    return count


def safe_replace(formula, old_text, new_text):
    """
    Given a raw spreadsheet formula, will replace all instances of old_text
    with new_text. 

    However, will not replace any of these instances _within_ strings
    within the formula itself.

    NOTE: this assumes the formula is valid!

    NOTE: old_text and new_text _must_ be words themselves, which means they are
    made up of A-Z, a-z, 0-9, and _.
    """

    if formula is None:
        return None

    # we do not edit inside the string literals! 
    string_ranges = get_string_ranges(formula)

    def replace(match):
        text: str = match.group()

        if match_in_strings(string_ranges, match):
            return text

        # Return the new text if it's the old text
        if text == old_text:
            return new_text

        return text
    
    # Perform the replacement
    return re.sub('\w+', replace, formula)


def check_common_errors(formula: str):
    """
    Helper function for checking a formula for common errors, for better
    communication with users. 

    If you want to throw a custom, specific error that is common for users 
    that can be detected when parsing a formula, add it here!
    """

    if safe_contains(formula, "<>"):
        raise make_invalid_formula_error(
            formula,
            'Please use != instead of <> to check inequality.'
        )

    # If the user used a lookup formula, point them to merge instead!
    LOOKUP_FORMULAS = ['VLOOKUP', 'HLOOKUP', 'XLOOKUP', 'LOOKUP']
    for lookup_formula in LOOKUP_FORMULAS:
        if safe_contains_function(formula.upper(), lookup_formula):
            raise make_invalid_formula_error(
                formula,
                f'Instead of {lookup_formula}, try using the merge button in the toolbar!'
            )
        
    if safe_count_function(formula, '\(') != safe_count_function(formula, '\)'):
        raise make_invalid_formula_error(
            formula,
            f'Sorry, it looks like your parentheses are uneven. Did you close them correctly?'
        )



def parse_formula(formula, address, throw_errors=True):
    """
    Returns a representation of the formula that is easy to handle.

    Specifically, this function returns the triple:
    (python_code, functions, dependencies).

    python_code : a string of Python code that executes
    this formula.
    functions : a set of a strings (function names) that 
    are called
    dependencies : a set of columns this formula references
    """
    if formula is None:
        return '', set(), set()

    # Check the formula for common errors, before parsing it
    # NOTE: this function throws errors!
    if throw_errors:
        check_common_errors(formula)

    # Chop off the =, if it exists. We also accept formulas
    # that don't have an equals
    if formula.startswith('='):
        formula = formula[1:]

    # we do not edit inside the string literals! 
    string_ranges = get_string_ranges(formula)

    functions = set()
    dependencies = set()
    def replace(match):
        """
        Each word match can be:
            1. A constant.
                - A number. Thus, all characters must be digits
                - A string. Must be surrounded by single or double quotes.
                - A boolean. True or False only.
            2. A function call. 
                - This MUST be followed by a '('
            3. A column_reference
                - Any word that isn't any of the above!
        """
        text: str = match.group()
        start = match.start()
        end = match.end() # this is +1 after the last char of the string

        # We skip this token if it is within a string!
        if match_in_strings(string_ranges, match):
            return text

        # CONSTANTS

        # Number
        if text.isnumeric():
            return text
        # String (check it's in quotes)
        if start - 1 >= 0 and (formula[start - 1] == '\"' or formula[start - 1] == '\'') \
            and end < len(formula) and (formula[end] == '\"' or formula[end] == '\''):
            return text
        # Boolean
        if text == 'True' or text == 'False':
            return text

        # Function
        if end < len(formula) and formula[end] == '(':
            # We turn all used functions into upper case in the translated Python
            # NOTE: this does not effect the original spreadsheet formula, which
            # may remaind lower case. 
            function = text.upper()
            functions.add(function)
            return function

        # Finially, columns
        dependencies.add(text)
        return f'df[\'{text}\']'
    
    # We match all words in formula, and send them through the replace function.
    # See documentation here: https://docs.python.org/3/library/re.html#re.sub
    formula = re.sub('\w+', replace, formula)
    
    # Finially, prepend the address to set the dataframe
    formula = f'df[\'{address}\'] = {formula}'

    return formula, functions, dependencies


"""
Schemas for the user.json file. As we evolve the format overtime,
we should include them here, so that we can track what upgrades
are required / necessary.

NOTE: USER_JSON_DEFAULT is the default user.json object for the currrent
version always.
"""

from datetime import datetime
from mitosheet.utils import get_random_id
from mitosheet._version import __version__

# Some helpful constants
GITHUB_ACTION_ID = 'github_action'
GITHUB_ACTION_EMAIL = 'github@action.com'

# Old field names
UJ_INTENDED_BEHAVIOR = 'intended_behavior'
UJ_CLOSED_FEEDBACK = 'closed_feedback'
UJ_MITOSHEET_LAST_FIVE_USAGES = 'mitosheet_last_five_usages'

# Current field names
UJ_USER_JSON_VERSION = 'user_json_version'
UJ_STATIC_USER_ID = 'static_user_id'
UJ_USER_SALT = 'user_salt'
UJ_USER_EMAIL = 'user_email'
UJ_RECEIVED_TOURS = 'received_tours'
UJ_FEEDBACKS = 'feedbacks'
UJ_MITOSHEET_CURRENT_VERSION = 'mitosheet_current_version'
UJ_MITOSHEET_LAST_UPGRADED_DATE = 'mitosheet_last_upgraded_date'
UJ_MITOSHEET_LAST_FIFTY_USAGES = 'mitosheet_last_fifty_usages'
UJ_MITOSHEET_TELEMETRY = 'mitosheet_telemetry'

# This was the final version of the user.json with user_json_version == 1
# and since we added fields to it over time, we make sure that all users 
# upgrading from version 1 actually get all the fields they need before 
# being upgraded
# NOTE: tests ensure we cannot change this object
USER_JSON_VERSION_1 = {
    UJ_USER_JSON_VERSION: 1,
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    UJ_USER_EMAIL: '',
    # A list of actions the user intends to do on the tool, which they fill
    # out when they sign up
    UJ_INTENDED_BEHAVIOR: [],
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    # If the user opted out of feedback, we store that they opted out, so
    # that we don't bombard them with feedback
    UJ_CLOSED_FEEDBACK: False,
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIVE_USAGES: [datetime.today().strftime('%Y-%m-%d')]
}

USER_JSON_VERSION_2 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 2,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')]
}

USER_JSON_VERSION_3 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 3,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True
}


# This is the most up to date user json, and you must update it when
# you add a new schema
USER_JSON_DEFAULT = USER_JSON_VERSION_3


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains functions for upgrading analyses!

NOTE: when adding a specific function for upgrading one version of a step to the other, the
format of the function should be:

def upgrade_<old_step_type>_<old_step_version>_to_<new_step_type>_<new_step_version>(step):
    ....
"""

from mitosheet.saved_analyses.schema_utils import upgrade_saved_analysis_format_to_steps_data
from mitosheet._version import __version__


def upgrade_group_1_to_pivot_2(step):
    """
    Upgrades from a group 1 step to a pivot 2 step, simply
    by changing the names of the params.

    Old format of the step: {
        "step_version": 1, 
        "step_type": "group", 
        "sheet_index": x, 
        "group_rows": [...], 
        "group_columns": [...], 
        "values": {...: ...}}}
    }

    New format of the step: {
        "step_version": 2, 
        "step_type": "pivot", 
        "sheet_index": old['sheet_index'], 
        "pivot_rows": old['group_rows'], 
        "pivot_columns": old['group_columns'], 
        "values": old['values']
    }
    """
    return {
        'step_version': 2,
        'step_type': 'pivot',
        'sheet_index': step['sheet_index'],
        'pivot_rows': step['group_rows'],
        'pivot_columns': step['group_columns'],
        'values': step['values']
    }
def upgrade_pivot_2_to_pivot_3(step):
    """
    Upgrades from a pivot 2 step to a pivot 3 step, which simple
    changes the formats of the values key to be a list rather than
    a single element

    Old format of the step: {
        "step_version": 2, 
        "step_type": "pivot", 
        "sheet_index": int, 
        "group_rows": string[], 
        "group_columns": string[], 
        "values": {string: AggregationType}
    }

    New format of the step: {
        "step_version": 3, 
        "step_type": "pivot", 
        "sheet_index": old['sheet_index'], 
        "pivot_rows": old['pivot_rows'], 
        "pivot_columns": old['pivot_columns'], 
        "values": {string: AggregationType[]} <- note the single item in this list is the previous single value
    }
    """
    return {
        'step_version': 3,
        'step_type': 'pivot',
        'sheet_index': step['sheet_index'],
        'pivot_rows': step['pivot_rows'],
        'pivot_columns': step['pivot_columns'],
        'values': {
            column_header: [aggregationType] for column_header, aggregationType in step['values'].items()
        }
    }


def upgrade_add_column_1_to_add_column_2(step): 
    """
    Upgrades from a add column 1 step to a add column 2 step, simply
    by adding the column_header_index param

    We just set the column_header_index to -1, so that it gets added to the 
    end of the dataframe, just like all previous analyses expect.

    Old format of the step: {
        "step_version": 1, 
        "step_type": "add_column", 
        'sheet_index', 
        'column_header'
    }

    New format of the step: {
        "step_version": 2, 
        "step_type": "add_column", 
        'sheet_index', 
        'column_header', 
        'column_header_index'
    }
    """
    return {
        "step_version": 2, 
        "step_type": "add_column", 
        'sheet_index': step['sheet_index'],
        'column_header': step['column_header'],
        'column_header_index': -1 # we set the column_header_index to -1 so that it gets added to the end
    }


def update_filter_column_1_to_filter_column_2(step): 
    """
    Upgrades from a filter column version 1 step to version 2. This was
    part of the refactor that unified type handling across the codebase,
    and got rid of an old distinction between a mito column type and a filter
    type. Now, we just have mito types!

    Old format of the step: {
        "step_version": 1, 
        "step_type": "filter_column", 
        "sheet_index": 0, 
        "column_header": "A", 
        "filters": [
            {"type": "number", "condition": "greater", "value": 1}, 
            {"filters": [
                {"type": "number", "condition": "greater", "value": 2}, 
                {"type": "number", "condition": "greater", "value": 3}], 
                "operator": "And"}
            ], 
        "operator": "And"}
    }

    New version of the step: {
        "step_version": 1, 
        "step_type": "filter_column", 
        "sheet_index": 0, 
        "column_header": "A", 
        "filters": [
            {"type": "number", "condition": "greater", "value": 1}, 
            {"filters": [
                {"type": "number", "condition": "greater", "value": 2}, 
                {"type": "number", "condition": "greater", "value": 3}], 
                "operator": "And"}
            ], 
        "operator": "And"}
    }

    Anywhere it says "type", we change from:
    - number -> number_series
    - string -> string_series
    - datetime -> datetime_series
    """
    type_mapping = {
        'string': 'string_series',
        'number': 'number_series',
        'datetime': 'datetime_series',
    }

    new_filters = []

    for filter_or_group in step['filters']:
        # We check if this is a filter or a group, and 
        # case appropriately!
        if 'filters' in filter_or_group:
            # Filter group case
            group = filter_or_group
            new_filters_in_group = []
            for filter_ in group['filters']:
                filter_['type'] = type_mapping[filter_['type']]
                new_filters_in_group.append(filter_)
            group['filters'] = new_filters_in_group

            new_filters.append(filter_or_group)
        else:
            # Just a regular filter, so rename for clarity
            filter_ = filter_or_group
            filter_['type'] = type_mapping[filter_['type']]
            new_filters.append(filter_)

    return {
        "step_version": 2, 
        "step_type": "filter_column", 
        "sheet_index": step['sheet_index'], 
        "column_header": step['column_header'], 
        "filters": new_filters, 
        "operator": step['operator']
    }

def upgrade_merge_1_to_merge_2(step): 
    """
    Upgrades from a merge 1 step to a merge 2 step, simply
    by adding the how param

    We just set the how param to 'lookup' 

    Old format of the step: {
        'step_version': 1, 
        'step_type': "merge", 
        'params': {
            'sheet_index_one': 0
            'merge_key_one': 'A'
            'selected_columns_one': ['A', 'B', 'C']
            'sheet_index_two': 1
            'merge_key_two': 'A'
            'selected_columns_two': ['A', 'D']
        }
    }

    New format of the step: {
        'step_version': 1, 
        'step_type': "merge",
        'params': {
            'how': 'lookup'
            'sheet_index_one': 0
            'merge_key_one': 'A'
            'selected_columns_one': ['A', 'B', 'C']
            'sheet_index_two': 1
            'merge_key_two': 'A'
            'selected_columns_two': ['A', 'D']
        }
    }
    """

    params = step['params']
    # Add the how field to the params object and set it to 'lookup'
    params['how'] = 'lookup'

    return {
        "step_version": 2, 
        "step_type": "merge", 
        "params": params
    }



"""
STEP_UPGRADES_FUNCTION_MAPPING mapping contains a mapping of all steps that need to be upgraded. A step
x at version y needs to be upgraded if STEP_UPGRADES[x][y] is defined, and in fact 
this mapping contains the function that can be used to do the upgrade!
 
NOTE: upgrades of steps should form a linear graph of upgrades to the most up to date
version. For example, if we change add_column from version 1 to version 2 to version 3, 
this object should contain:
    {
        'add_column': {
            1: upgrade_add_column_1_to_add_column_2,
            2: upgrade_add_column_2_to_add_column_3
        }
    }
"""
STEP_UPGRADES_FUNCTION_MAPPING_OLD_FORMAT = {
    'group': {
        1: upgrade_group_1_to_pivot_2
    }, 
    'pivot': {
        2: upgrade_pivot_2_to_pivot_3
    },
    'add_column': {
        1: upgrade_add_column_1_to_add_column_2
    },
    'filter_column': {
        1: update_filter_column_1_to_filter_column_2
    }
}

STEP_UPGRADES_FUNCTION_MAPPING_NEW_FORMAT = {
    'merge': {
        1: upgrade_merge_1_to_merge_2
    }
}



def is_prev_version(version: str, curr_version: str=__version__):
    """
    Returns True if the passed version is a previous version compared
    to the current version; note that this assumes semantic versioning
    with x.y.z!
    """
    old_version_parts = version.split('.')
    curr_version_parts = curr_version.split('.')

    for old_version_part, curr_version_part in zip(old_version_parts, curr_version_parts):
        if int(old_version_part) > int(curr_version_part):
            # E.g. if we have 0.2.11 and 0.1.11, we want to return early as it's clearly not older!
            return False

        if int(old_version_part) < int(curr_version_part):
            return True

    return False


def upgrade_step_to_current(step, step_upgrade_function_mapping):
    """
    A recursive helper function that, given a step, keeps upgrading it
    until it cannot be upgraded anymore

    Takes a step_upgrade_function_mapping as input, which can either be 
    STEP_UPGRADES_FUNCTION_MAPPING_OLD_FORMAT or STEP_UPGRADES_FUNCTION_MAPPING_NEW_FORMAT
    """
    step_version = step['step_version']
    step_type = step['step_type']

    if step_type in step_upgrade_function_mapping:
        step_upgrades = step_upgrade_function_mapping[step_type]
        if step_version in step_upgrades:
            return upgrade_step_to_current(step_upgrades[step_version](step), step_upgrade_function_mapping)
    
    return step

def upgrade_steps_for_old_format(saved_analysis):
    """
    A helper function that operates on the old analysis format of:
    {
        "version": "0.1.197",
        "steps": {
            1: ...
        }
    }

    And makes sure all their steps are up to date.
    """
    if saved_analysis is None:
        return None

    # If it's already in the new format, then don't worry about it
    if 'steps' not in saved_analysis and 'steps_data' in saved_analysis:
        return saved_analysis
        
    version = saved_analysis["version"]
    steps = saved_analysis["steps"]

    if not is_prev_version(version):
        # TODO: do we want to _change_ the version this was saved with? I think it doesn't
        # really matter, as it gets changed when it gets rewritten...
        return saved_analysis

    new_steps = []
    for step_idx, step in steps.items():
        new_steps.append(upgrade_step_to_current(step, STEP_UPGRADES_FUNCTION_MAPPING_OLD_FORMAT))

    # Convert the new steps in the correct format
    new_steps_json = {
        str(i + 1): step for i, step in enumerate(new_steps)
    }

    return {
        'version': __version__,
        'steps': new_steps_json
    }

def upgrade_steps_for_new_format(saved_analysis):
    """
    A helper function that operates on the new analysis format of:

    "version": "0.1.197",
    "steps_data": [
        {
            "step_version": 1,
            "step_type": "filter",
            "params": {
                ....
            }
        }
    ]

    And makes sure all their steps are up to date.
    """
    if saved_analysis is None:
        return None
        
    version = saved_analysis["version"]
    steps_data = saved_analysis["steps_data"]

    new_steps_data = []
    for step in steps_data:
        new_steps_data.append(upgrade_step_to_current(step, STEP_UPGRADES_FUNCTION_MAPPING_NEW_FORMAT))

    return {
        'version': __version__,
        'steps_data': new_steps_data
    }

def upgrade_saved_analysis_to_current_version(saved_analysis):
    """
    Upgrades a saved analysis to the current version.
    
    Notable, changes to the saved analysis take two types:
    1. Changes to the format of the saved analysis itself. 
    2. Changes to the format of the specific steps in the saved analysis.

    See mitosheet/upgrade/schemas.py, but as we only have 1 format change
    in the saved analyses, we process the specific step upgrades first if
    they exist.
    """
    saved_analysis = upgrade_steps_for_old_format(saved_analysis)
    new_format_saved_analysis = upgrade_saved_analysis_format_to_steps_data(saved_analysis)
    saved_analysis = upgrade_steps_for_new_format(new_format_saved_analysis)

    return saved_analysis
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.


from copy import copy
from mitosheet.state import DATAFRAME_SOURCE_PIVOTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from typing import Any, Dict, List, Optional
from mitosheet.utils import make_valid_header
import pandas as pd
from pandas.core.base import DataError
import pandas as pd

from mitosheet.errors import (
    make_invalid_pivot_error,
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_aggregation_error,
)

# Aggregation types pivot supports
PA_COUNT_UNIQUE = 'count unique'
PIVOT_AGGREGATION_TYPES = [
    # These first few are supported out of the box by 
    # pandas, so we don't need any extra support for them
    'sum',
    'mean',
    'median',
    'min',
    'max', 
    'count', 
    'std',
    PA_COUNT_UNIQUE
]

class PivotStepPerformer(StepPerformer):
    """
    A pivot, which allows you to pivot data from an existing dataframe 
    along some keys, and then aggregate columns with specific functions.
    """

    @classmethod
    def step_version(cls) -> int:
        return 3

    @classmethod
    def step_type(cls) -> str:
        return 'pivot'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Pivoted a Dataframe'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'pivot_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        """
        Saturates the pivot table with just a `created_non_empty_dataframe` key, which
        is useful for logging.

        Furthermore, we filter out any duplicated aggregation keys, as they
        result in errors without adding any data to the pivot.
        """
        # Case 1 - we have at least one row and at least one value
        created_case_1 = len(params['pivot_rows']) > 0 and len(params['values']) > 0
        # Case 2 - we have at least one column and at least one value
        created_case_2 = len(params['pivot_columns']) > 0 and len(params['values']) > 0
        params['created_non_empty_dataframe'] = created_case_1 or created_case_2

        # Filter out any duplicate aggregation functions
        for column_header, aggregation_function_names in params['values'].items():
            seen = set()
            aggregation_function_names = [x for x in aggregation_function_names if not (x in seen or seen.add(x))]
            params['values'][column_header] = aggregation_function_names

        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        pivot_rows: List[str],
        pivot_columns: List[str],
        values: Dict[str, List[str]],
        destination_sheet_index: int=None,
        **params
    ) -> State:

        # if the sheets don't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error(sheet_index)

        # We check that the pivot by doesn't use any columns that don't exist
        columns_used = set(pivot_rows).union(set(pivot_columns)).union(set(values.keys()))
        missing_pivot_keys = columns_used.difference(prev_state.column_metatype[sheet_index].keys())
        if len(missing_pivot_keys) > 0:
            raise make_no_column_error(missing_pivot_keys)

        # Create the post state, it can be a shallow copy
        post_state = copy(prev_state)

        try:
            # Actually execute the pivoting
            new_df = _execute_pivot(
                prev_state.dfs[sheet_index], 
                pivot_rows,
                pivot_columns,
                values
            )
            
            destination_sheet_index = post_state.add_df_to_state(
                new_df, 
                DATAFRAME_SOURCE_PIVOTED,
                sheet_index=destination_sheet_index
            )

        except DataError as e:
            # A data-error occurs when you try to aggregate on a column where the function
            # cannot be applied (e.g. 'mean' on a column of strings)
            print(e)
            # Generate an error informing the user
            raise make_invalid_aggregation_error()


        return post_state, {
            'destination_sheet_index': destination_sheet_index
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        pivot_rows: List[str],
        pivot_columns: List[str],
        values: Dict[str, List[str]],
        destination_sheet_index: int=None,
        **params
    ) -> List[str]:
        old_df_name = post_state.df_names[sheet_index]
        if destination_sheet_index is None:
            new_df_name = f'df{len(post_state.dfs)}'
        else:
            # If we're repivoting an existing pivot table, we have
            # to make sure to overwrite the correct pivot table 
            # by using the right name
            new_df_name = post_state.df_names[destination_sheet_index]
        
        # If there are no keys or values to aggregate on we return an empty dataframe. 
        if len(pivot_rows) == 0 and len(pivot_columns) == 0 or len(values) == 0:
            return [f'{new_df_name} = pd.DataFrame(data={{}})']

        transpiled_code = []

        # Drop any columns we don't need, to avoid issues where pandas freaks out
        # and says there is a non-1-dimensional grouper
        transpiled_code.append(f'unused_columns = {old_df_name}.columns.difference(set({pivot_rows}).union(set({pivot_columns})).union(set({set(values.keys())})))')
        transpiled_code.append(f'tmp_df = {old_df_name}.drop(unused_columns, axis=1)')

        # Do the actual pivot
        pivot_table_args = build_args_code(pivot_rows, pivot_columns, values)
        transpiled_code.append(f'pivot_table = tmp_df.pivot_table({NEWLINE_TAB}{pivot_table_args}\n)')

        # Flatten column headers, which we always do because it's hard to tell when we should
        transpiled_code.append(FLATTEN_COMMENT)
        transpiled_code.append(FLATTEN_CODE)

        # Finially, reset the column name, and the indexes!
        transpiled_code.append(RESET_INDEX_COMMENT)
        transpiled_code.append(f'{new_df_name} = pivot_table.rename_axis(None, axis=1).reset_index()')

        return transpiled_code

    @classmethod
    def describe(
        cls,
        sheet_index,
        pivot_rows,
        pivot_columns,
        values,
        destination_sheet_index=None,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            new_df_name = f'df{len(df_names)}'
            old_df_name = df_names[sheet_index]
            return f'Pivoted {old_df_name} into {new_df_name}'
        return f'Pivoted dataframe {sheet_index}'


def values_to_functions(values):
    """
    Helper function for turning the values mapping sent by the frontend to the values
    mapping to functions that can actually be passed to the pandas pivot function
    """
    new_values = dict()

    for column_header, aggregation_function_names in values.items():
        new_values[column_header] = []
        for agg_func_name in aggregation_function_names:
            if agg_func_name not in PIVOT_AGGREGATION_TYPES:
                raise make_invalid_pivot_error()

            if agg_func_name == PA_COUNT_UNIQUE:
                agg_func = pd.Series.nunique
            else:
                agg_func = agg_func_name
                
            new_values[column_header].append(agg_func)
    
    return new_values

def _execute_pivot(
        df: pd.DataFrame, 
        pivot_rows: List[str], 
        pivot_columns: List[str], 
        values
    ):
    """
    Helper function for executing the pivot on a specific dataframe
    and then aggregating the values with the passed values mapping
    """

    # If there are no keys to aggregate on, we return an empty dataframe
    if (len(pivot_rows) == 0 and len(pivot_columns) == 0) or len(values) == 0:
        return pd.DataFrame(data={})

    values_keys = list(values.keys())

    # Built the args, leaving out any unused values
    args = {}

    if len(pivot_rows) > 0:
        args['index'] = pivot_rows

    if len(pivot_columns) > 0:
        args['columns'] = pivot_columns

    if len(values) > 0:
        args['values'] = values_keys
        args['aggfunc'] = values_to_functions(values)


    # Before execution, we make a temp dataframe that does not have the columns 
    # we do not need, as this allows us to avoid a bug in pandas where these extra
    # columns cause a data
    unused_columns = df.columns.difference(set(pivot_rows).union(set(pivot_columns)).union(set(values_keys)))
    df = df.drop(unused_columns, axis=1)

    # Actually perform the pivot
    pivot_table = df.pivot_table(**args) # type: pd.DataFrame

    # Flatten the column headers
    pivot_table.columns = [make_valid_header(col) for col in pivot_table.columns.values]

    # flatten the column headers & reset the indexes
    pivot_table = pivot_table.rename_axis(None, axis=1).reset_index()

    return pivot_table

def values_to_functions_code(values):
    """
    Helper function for turning the values mapping sent by the frontend to the values
    mapping that works in generated code. Namely, needs to replay Count Unique with the
    function pd.Series.nunique.
    """
    string_values = f'{values}'
    # NOTE: this needs to match the values sent from the frontend
    # also note that we overwrite the quotes around Count Unique
    return string_values.replace('\'count unique\'', 'pd.Series.nunique')

# Helpful constants for code formatting
TAB = '    '
NEWLINE_TAB = f'\n{TAB}'

FLATTEN_COMMENT = '# Flatten the column headers'
FLATTEN_CODE = f'pivot_table.columns = [make_valid_header(col) for col in pivot_table.columns.values]'
RESET_INDEX_COMMENT = '# Reset the column name and the indexes'


def build_args_code(
        pivot_rows,
        pivot_columns,
        values
    ):
    """
    Helper function for building an arg string, while leaving
    out empty arguments. 
    """
    values_keys = list(values.keys())

    args = []
    if len(pivot_rows) > 0:
        args.append(f'index={pivot_rows},')

    if len(pivot_columns) > 0:
        args.append(f'columns={pivot_columns},')

    if len(values) > 0:
        args.append(f'values={values_keys},')
        args.append(f'aggfunc={values_to_functions_code(values)}')
        
    return NEWLINE_TAB.join(args)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pytest
from copy import deepcopy

from ..errors import EditError
from ..topological_sort import topological_sort_columns, creates_circularity

# Test circularity detection works

def test_creates_circularity_self_reference():
    column_evaluation_graph = {'A': set([]), 'B': set([]), 'C': set([])}
    should_not_modify = deepcopy(column_evaluation_graph)
    assert creates_circularity(should_not_modify, 'A', [], ['A'])
    assert column_evaluation_graph == should_not_modify

def test_creates_circularity_intermediate_reference():
    column_evaluation_graph = {'A': set(['B']), 'B': set(['C']), 'C': set([])}
    should_not_modify = deepcopy(column_evaluation_graph)
    assert creates_circularity(should_not_modify, 'A', [], ['C'])
    assert column_evaluation_graph == should_not_modify

def test_creates_no_circularity_intermediate_linear():
    column_evaluation_graph = {'A': set(['B']), 'B': set([]), 'C': set([])}
    should_not_modify = deepcopy(column_evaluation_graph)
    assert not creates_circularity(should_not_modify, 'C', [], ['B'])
    assert column_evaluation_graph == should_not_modify

# Test toplogical sort itself

def test_no_cell_relationships_no_error():
    column_evaluation_graph = {'A': set([]), 'B': set([]), 'C': set([])}
    topological_sort_columns(column_evaluation_graph)

def test_simple_linear_topological_sort():
    column_evaluation_graph = {'A': set(['B']), 'B': set(['C']), 'C': set([])}
    sort = topological_sort_columns(column_evaluation_graph)
    assert sort == ['A', 'B', 'C']

def test_simple_linear_topological_sort_options():
    column_evaluation_graph = {'A': set(['B']), 'B': set([]), 'C': set([])}
    sort = topological_sort_columns(column_evaluation_graph)
    assert (sort == ['A', 'B', 'C'] or sort == ['A', 'C', 'B'] or sort == ['C', 'A', 'B'])

def test_out_of_order_topological_sort_options():
    column_evaluation_graph = {
        'A': set([]), 
        'B': set(['A']), 
        'C': set(['A'])
    }
    sort = topological_sort_columns(column_evaluation_graph)
    assert (sort == ['C', 'B', 'A'] or sort == ['B', 'C', 'A'])

def test_out_of_order_linear_topological_sort_options():
    column_evaluation_graph = {
        'A': set([]), 
        'B': set(['A']), 
        'C': set(['B'])
    }
    sort = topological_sort_columns(column_evaluation_graph)
    assert sort == ['C', 'B', 'A']

def test_strong_linear_order():
    column_evaluation_graph = {
        'A': set(['B', 'C']), 
        'B': set(['C']), 
        'C': set([])
    }
    sort = topological_sort_columns(column_evaluation_graph)
    assert sort == ['A', 'B', 'C']

# Errors with cyles

def test_simple_cycle_is_cycle_error():
    column_evaluation_graph = {'A': set(['B']), 'B': set(['C']), 'C': set(['A'])}
    with pytest.raises(EditError) as edit_error_info:
        sort = topological_sort_columns(column_evaluation_graph)
    assert edit_error_info.value.type_ == 'circular_reference_error'

def test_self_reference_is_cycle_error():
    column_evaluation_graph = {'A': set(['B']), 'B': set(['C']), 'C': set(['C'])}
    with pytest.raises(EditError) as edit_error_info:
        sort = topological_sort_columns(column_evaluation_graph)
    assert edit_error_info.value.type_ == 'circular_reference_error'

def test_simple_nested_cycle_is_cycle_error():
    column_evaluation_graph = {'A': set(['B']), 'B': set(['C']), 'C': set(['B'])}
    with pytest.raises(EditError) as edit_error_info:
        sort = topological_sort_columns(column_evaluation_graph)
    assert edit_error_info.value.type_ == 'circular_reference_error'



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pytest

from mitosheet.errors import EditError
from mitosheet.parser import parse_formula, safe_replace, safe_contains

CONSTANT_TEST_CASES = [
    (
        '=100',
        'B',
        'df[\'B\'] = 100',
        set([]),
        set([])
    ),
    (
        '=True',
        'B',
        'df[\'B\'] = True',
        set([]),
        set([])
    ),
    (
        '=\'StringSingleQuotes\'',
        'B',
        'df[\'B\'] = \'StringSingleQuotes\'',
        set([]),
        set([])
    ),
    (
        '=\"StringDoubleQuotes\"',
        'B',
        'df[\'B\'] = \"StringDoubleQuotes\"',
        set([]),
        set([])
    ), 
    (
        '=\"String Double Quotes\"',
        'B',
        'df[\'B\'] = \"String Double Quotes\"',
        set([]),
        set([])
    ), 
    (
        '=FUNC(\"String Double Quotes\")',
        'B',
        'df[\'B\'] = FUNC(\"String Double Quotes\")',
        set(['FUNC']),
        set([])
    ),
    (
        '=FUNC(\"DIFF_FUNC(A)\")',
        'B',
        'df[\'B\'] = FUNC(\"DIFF_FUNC(A)\")',
        set(['FUNC']),
        set([])
    ),
    (
        '=\"String One FUNC(A)\" + \"STRING TWO FUNC(A)\"',
        'B',
        'df[\'B\'] = \"String One FUNC(A)\" + \"STRING TWO FUNC(A)\"',
        set([]),
        set([])
    ),
    (
        '=\'String One FUNC(A)\' + \'STRING TWO FUNC(A)\'',
        'B',
        'df[\'B\'] = \'String One FUNC(A)\' + \'STRING TWO FUNC(A)\'',
        set([]),
        set([])
    ),
    (
        '=FUNC1(\'String One FUNC(A)\') + FUNC2(\'STRING TWO FUNC(A)\')',
        'B',
        'df[\'B\'] = FUNC1(\'String One FUNC(A)\') + FUNC2(\'STRING TWO FUNC(A)\')',
        set(['FUNC1', 'FUNC2']),
        set([])
    )
]

# Tests cases to ensure operators are parsed correctly
OPERATOR_TEST_CASES = [
    # Simple add operation
    (
        '=A + B',
        'C',
        'df[\'C\'] = df[\'A\'] + df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple subtract operation
    (
        '=A - B',
        'C',
        'df[\'C\'] = df[\'A\'] - df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple multiply operation
    (
        '=A * B',
        'C',
        'df[\'C\'] = df[\'A\'] * df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple divide operation
    (
        '=A / B',
        'C',
        'df[\'C\'] = df[\'A\'] / df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Handles multi-char columns
    (
        '=AAA + BBB + CCC + DDD',
        'E',
        'df[\'E\'] = df[\'AAA\'] + df[\'BBB\'] + df[\'CCC\'] + df[\'DDD\']',
        set(),
        set(['AAA', 'BBB', 'CCC', 'DDD'])
    ),
    # Maintains parens and constants
    (
        '=(A + B) / C + A * 100',
        'D',
        'df[\'D\'] = (df[\'A\'] + df[\'B\']) / df[\'C\'] + df[\'A\'] * 100',
        set(),
        set(['A', 'B', 'C'])
    ),
    # Operator in functions
    (
        '=FUNC(A + B / C + A * 100)',
        'D',
        'df[\'D\'] = FUNC(df[\'A\'] + df[\'B\'] / df[\'C\'] + df[\'A\'] * 100)',
        set(['FUNC']),
        set(['A', 'B', 'C'])
    )
]

# Tests proper function parsing
FUNCTION_TEST_CASES = [
    # Simple, singular function call
    (
        '=FUNC(A)',
        'B',
        'df[\'B\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Simple, singular function call that has the name of a column
    (
        '=C(A)',
        'B',
        'df[\'B\'] = C(df[\'A\'])',
        set(['C']),
        set(['A'])
    ),
    # Nested functions with different names
    (
        '=FUNCA(FUNCB(FUNCC(FUNCD(FUNCE(FUNCF(FUNCG(FUNCH(A))))))))',
        'B',
        'df[\'B\'] = FUNCA(FUNCB(FUNCC(FUNCD(FUNCE(FUNCF(FUNCG(FUNCH(df[\'A\']))))))))',
        set(['FUNCA', 'FUNCB', 'FUNCC', 'FUNCD', 'FUNCE', 'FUNCF', 'FUNCG', 'FUNCH']),
        set(['A'])
    ),
    # Nested functions with basic operators
    (
        '=RIGHT(A, LEN(B) - 10)',
        'C',
        'df[\'C\'] = RIGHT(df[\'A\'], LEN(df[\'B\']) - 10)',
        set(['RIGHT', 'LEN']),
        set(['A', 'B'])
    ),
    # More nested functions with columns
    (
        '=FUNC(A, FUNC(B, FUNC(C)))',
        'D',
        'df[\'D\'] = FUNC(df[\'A\'], FUNC(df[\'B\'], FUNC(df[\'C\'])))',
        set(['FUNC']),
        set(['A', 'B', 'C'])
    ),
    # Lots of column references
    (
        '=FUNC(A, B, C, D, E, F, G, H, I)',
        'J',
        'df[\'J\'] = FUNC(df[\'A\'], df[\'B\'], df[\'C\'], df[\'D\'], df[\'E\'], df[\'F\'], df[\'G\'], df[\'H\'], df[\'I\'])',
        set(['FUNC']),
        set(['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'])
    ),
    # Lots of column references to columns with numbers
    (
        '=FUNC(column1, column2)',
        'J',
        'df[\'J\'] = FUNC(df[\'column1\'], df[\'column2\'])',
        set(['FUNC']),
        set(['column1', 'column2'])
    ),
    # Test references to columns with underscores
    (
        '=FUNC(column_1, column_2)',
        'J',
        'df[\'J\'] = FUNC(df[\'column_1\'], df[\'column_2\'])',
        set(['FUNC']),
        set(['column_1', 'column_2'])
    ),
    # Test references to columns that are lookups
    (
        '=FUNC(LOOKUP, XLOOKUP, VLOOKUP, HLOOKUP)',
        'J',
        'df[\'J\'] = FUNC(df[\'LOOKUP\'], df[\'XLOOKUP\'], df[\'VLOOKUP\'], df[\'HLOOKUP\'])',
        set(['FUNC']),
        set(['LOOKUP', 'XLOOKUP', 'VLOOKUP', 'HLOOKUP'])
    ) 
]


"""
PARSE_TESTS contains a variety of tests that make sure
formula parsing is working properly; it is passed as 
a parameter into the test_parse test below.

Order of params is: formula, address, python_code, functions, columns

See documentation here: https://docs.pytest.org/en/latest/parametrize.html#parametrize-basics
"""
PARSE_TESTS = CONSTANT_TEST_CASES + OPERATOR_TEST_CASES + FUNCTION_TEST_CASES
@pytest.mark.parametrize("formula,address,python_code,functions,columns", PARSE_TESTS)
def test_parse(formula, address, python_code, functions, columns):
    assert parse_formula(formula, address) == \
        (
            python_code, 
            functions, 
            columns
        )


PARSE_TEST_ERRORS = [
    ('=LOOKUP(100, A)', 'B', 'invalid_formula_error', 'LOOKUP'),
    ('=VLOOKUP(100, A)', 'B', 'invalid_formula_error', 'VLOOKUP'),
    ('=HLOOKUP(100, A)', 'B', 'invalid_formula_error', 'HLOOKUP'),
    ('=XLOOKUP(100, A)', 'B', 'invalid_formula_error', 'XLOOKUP'),
    ('=A <> 100', 'B', 'invalid_formula_error', '<>'),
    ('=SUM(A', 'B', 'invalid_formula_error', 'parentheses'),
]
@pytest.mark.parametrize("formula, address, error_type, to_fix_substr", PARSE_TEST_ERRORS)
def test_parse_errors(formula, address, error_type, to_fix_substr):
    with pytest.raises(EditError) as e_info:
        parse_formula(formula, address)
    assert e_info.value.type_ == error_type
    if to_fix_substr is not None:
        assert to_fix_substr in e_info.value.to_fix


SAFE_REPLACE_TESTS = [
    ('=A', 'A', 'B', '=B'),
    ('=A + A', 'A', 'B', '=B + B'),
    ('=A + B', 'A', 'B', '=B + B'),
    ('=A + B + A', 'A', 'B', '=B + B + B'),
    ('=A + B + \"A\"', 'A', 'B', '=B + B + \"A\"'),
    ('=A + B + \'A\'', 'A', 'B', '=B + B + \'A\''),
    ('=A + B + \'A\'', 'A', 'B', '=B + B + \'A\''),
    ('=FUNC(A, B, A) + TEST(FUNC(A, \'A\', \'B\')) + \'A\'', 'A', 'B', '=FUNC(B, B, B) + TEST(FUNC(B, \'A\', \'B\')) + \'A\''),
    ('=APPLE(A, B, A) + AARON(AAA(A, \'A\', \'B\')) + \'A\'', 'A', 'B', '=APPLE(B, B, B) + AARON(AAA(B, \'A\', \'B\')) + \'A\''),
]

@pytest.mark.parametrize('formula,old_text,new_text,new_formula', SAFE_REPLACE_TESTS)
def test_safe_replace(formula, old_text, new_text, new_formula):
    assert safe_replace(formula, old_text, new_text) == new_formula


SAFE_CONTAINS_TESTS = [
    ('=A', 'A', True),
    ('=A + B', 'B', True),
    ('=A + "B"', 'B', False),
    ('=A & "B"', '&', True),
    ('=A - +  "&"', '&', False),
    ('=A - +  \'&\'', '&', False),
]

@pytest.mark.parametrize('formula,substring,contains', SAFE_CONTAINS_TESTS)
def test_safe_contains(formula, substring, contains):
    assert safe_contains(formula, substring) == contains


import json
from mitosheet.step_performers.pivot import PivotStepPerformer



def get_pivot_params(event, steps_manager):
    """
    Returns the 'params' object for a pivot step that pivoted
    into this sheet at sheet_index. Notably, returns the _last_
    params that were for this pivot.
    """
    destination_sheet_index = event['destination_sheet_index']

    # Loop over the steps backwards, so that we get the most recent one
    for step in steps_manager.steps[:steps_manager.curr_step_idx + 1][::-1]:
        if step.step_type == PivotStepPerformer.step_type() \
            and step.execution_data['destination_sheet_index'] == destination_sheet_index:
            return json.dumps({
                'sheet_index': step.params['sheet_index'],
                'pivot_rows': step.params['pivot_rows'],
                'pivot_columns': step.params['pivot_columns'],
                'values': step.params['values'],
            })
    
    # Return nothing, if there is no pivot that meets this criteria
    return ''


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import copy
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.state import State
from typing import Any, Dict, List, Optional


class DataframeDeleteStepPerformer(StepPerformer):
    """
    Deletes a dataframe from everywhere in the step.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'dataframe_delete'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Deleted a Dataframe'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'dataframe_delete_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        sheet_index = params['sheet_index']
        old_dataframe_name = prev_state.df_names[sheet_index]
        params['old_dataframe_name'] = old_dataframe_name
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        old_dataframe_name: str,
        **params
    ) -> State:
        # Create a new step and save the parameters
        post_state = copy(prev_state)

        # Execute the delete
        post_state.column_metatype.pop(sheet_index)
        post_state.column_type.pop(sheet_index)
        post_state.column_spreadsheet_code.pop(sheet_index)
        post_state.column_python_code.pop(sheet_index)
        post_state.column_evaluation_graph.pop(sheet_index)
        post_state.column_filters.pop(sheet_index)
        post_state.dfs.pop(sheet_index)
        post_state.df_names.pop(sheet_index)
        post_state.df_sources.pop(sheet_index)

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        old_dataframe_name: str
    ) -> List[str]:
        return [f'del {old_dataframe_name}']

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        old_dataframe_name: str,
        df_names=None,
        **params
    ) -> str:
        return f'Deleted dataframe {old_dataframe_name}'


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
For going to a number series.
"""
from typing import Tuple, Union
import pandas as pd
import numpy as np

from mitosheet.sheet_functions.types.utils import BOOLEAN_SERIES, DATETIME_SERIES, NUMBER_SERIES, STRING_SERIES, get_mito_type


def convert_string_to_number(
        s, 
        on_uncastable_arg_element, #Union[Literal['error'], Tuple[Literal['default'], any]]
    ):
    """
    NOTE: The approach for this function is to start as a string, and then we try
    and turn each element into a number 1-by-1. We attempt to handle:
    1. All basic integers / floats.
    2. Dollar signs. 
    3. Parenthases to denote negative numbers, per accounting conventions. 
    4. Commas in the string (e.g. 123,456 => 123456). NOTE: We handle the European conventions here too, 
        where commas may represent a decimal. However, this is impossible to make perfect, as 123,123 and
        123,123 are 123123 and 123.123 in America and Europe respectively. We treat this as American for 
        now. 
    As these are all heuristics, we do our best. We also try to perform this conversion
    optimistically, as to run as quickly as possible.
    """
    try:
        # Try to handle case 1, optimistically
        return float(s)
    except:
        # Get rid of whitespace at the ends
        s = s.strip()

        is_negative = False
        if s.startswith('-'):
            s = s[1:]
            is_negative = True

        # Handle 2, if it exist
        if s.startswith('$'):
            s = s[1:]

        # Handle 3, if it exists
        if s.startswith('(') and s.endswith(')'):
            s = s[1:-1]
            is_negative = True

        # Finially, handle case 4, if it's an issue
        if ',' in s:
            # We try and figure out if it's a european or american comma usage, by seeing
            # what happens at the end of the string
            last_comma_index = s.rfind(',')

            # If there is no period, and the the string after the last comma is anything other
            # than 3 characters long, than we take this as a European, and turn the comma into a .
            if '.' not in s and last_comma_index != len(s) - 4:
                s = s.replace(',', '.')
            else:
                # Otherwise, we treat this as American
                s = s.replace(',', '')
    
        try:
            return float(s) * (-1 if is_negative else 1)
        except:
            if on_uncastable_arg_element == 'error':
                # TODO: raise a better exception
                raise Exception('Invalid element')
            else:
                # Return the given default value in this case
                return on_uncastable_arg_element[1]

def to_number_series_from_string_series(
        string_series: pd.Series, 
        on_uncastable_arg_element #: Union[Literal['error'], Tuple[Literal['default'], any]]
    ):
    """
    Converts a string series to a number series, using a helper that
    handles special formatting of strings. 

    Takes a default value, so the tranformation can occur elementwise
    """
    return string_series.apply(convert_string_to_number, on_uncastable_arg_element=on_uncastable_arg_element).astype('float64')

def to_number_series_from_boolean_series(boolean_series: pd.Series):
    """
    As all boolean series are very easily convertable to a number_series, 
    and cannot fail on any element, so we can do this in one easy move.

    Note that False -> 0, and True -> 1.
    """
    return boolean_series.astype('float64')


def to_number_series(
        unknown_object,
        on_uncastable_arg_element=('default', np.NaN), # Union[Literal['error'], Tuple[Literal['default'], any]]
    ):
    from_type = get_mito_type(unknown_object)

    # If it is not a series, we put it in a series, and get the type again
    if not from_type.endswith('series'):
        unknown_object = pd.Series([unknown_object])
        from_type = get_mito_type(unknown_object)

    if from_type == BOOLEAN_SERIES:
        return to_number_series_from_boolean_series(unknown_object)
    elif from_type == DATETIME_SERIES:
        return None
    elif from_type == NUMBER_SERIES:
        return unknown_object
    elif from_type == STRING_SERIES:
        return to_number_series_from_string_series(unknown_object, on_uncastable_arg_element=on_uncastable_arg_element)
    else:
        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains functions that are useful for determining the state of the
current user.
"""
import os
import getpass
from datetime import datetime

from mitosheet._version import __version__
from mitosheet.user.db import get_user_field
from mitosheet.user.schemas import (
    UJ_FEEDBACKS, UJ_MITOSHEET_LAST_UPGRADED_DATE, UJ_MITOSHEET_LAST_FIFTY_USAGES
)

def is_running_test():
    """
    A helper function that quickly returns if the current code is running 
    inside of a test, which is useful for making sure we don't generate 
    tons of logs.
    """
    # Pytest injects PYTEST_CURRENT_TEST into the current enviornment when running
    running_pytests = "PYTEST_CURRENT_TEST" in os.environ
    # Github injects CI into the enviornment when running
    running_ci = 'CI' in os.environ and os.environ['CI'] is not None

    return running_pytests or running_ci

def is_on_kuberentes_mito():
    """
    Returns True if the user is on Kuberentes Mito, on staging or on app
    """
    user = getpass.getuser()
    return user == 'jovyan'


def is_local_deployment():
    """
    Helper function for figuring out if this a local deployment or a
    Mito server deployment
    """
    return not is_on_kuberentes_mito()  


def should_upgrade_mitosheet():
    """
    A helper function that calculates if a user should upgrade, which does so by 
    checking if the user has upgraded in the past 21 days (3 weeks), since this is
    about how often we release big features.

    Always returns false if it is not a local installation, for obvious reasons.

    NOTE: if the user clicks the upgrade button in the app, then we change the upgraded 
    date to this date, so that the user doesn't get a bunch of annoying popups. This just
    pushes back when they are annoyed to upgrade!
    """
    if not is_local_deployment():
        return False

    mitosheet_last_upgraded_date = datetime.strptime(get_user_field(UJ_MITOSHEET_LAST_UPGRADED_DATE), '%Y-%m-%d')
    return (datetime.now() - mitosheet_last_upgraded_date).days > 21




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
This preprocessor reads in any arguments that are
strings, treats them as file paths, and attempts
to read the in a dataframes.
"""
from mitosheet.step_performers.import_steps.simple_import import get_dataframe_names
from mitosheet.errors import get_recent_traceback_as_list
import pandas as pd
from mitosheet.mito_analytics import log

def get_string_args(args):
    return [arg for arg in args if isinstance(arg, str)]

def execute_preprocess_read_file_paths(args):

    df_args = []
    for arg in args:
        if isinstance(arg, pd.DataFrame):
            df_args.append(arg)
        elif isinstance(arg, str):
            # If it is a string, we try and read it in as a dataframe
            try:
                df_args.append(
                    pd.read_csv(arg)
                )
            except:
                # If this pd.read_csv fails, then we report this error to the user
                # as a failed mitosheet call
                error_message = f'Invalid argument passed to sheet: {arg}. This path could not be read with a pd.read_csv call. Please pass in the parsed dataframe directly.'
                log('mitosheet_sheet_call_failed', {'error': error_message, 'error_traceback': get_recent_traceback_as_list()})
                raise ValueError(error_message)
        else:
            error_message = f'Invalid argument passed to sheet: {arg}. Please pass all dataframes or paths to CSV files.'
            log('mitosheet_sheet_call_failed', {'error': error_message})
            raise ValueError(error_message)
            
    return df_args

def transpile_preprocess_read_file_paths(
        steps_manager
    ):
    """
    Transpiles the reading in of passed file paths to dataframe names, 
    with a simple pd.read_csv call.
    """
    code = []

    # First, we get all the string arguments passed to the sheet call
    str_args = get_string_args(steps_manager.original_args)
    
    # Then, we turn these into dataframe names.
    # NOTE: there is potentially a bug if a user passes in a dataframe
    # with the same name as the the result of the return of this function,
    # but we ignore this for now. E.g. mitosheet.sheet('df1.csv', 'df1_csv') 
    # will cause only variable to exist.
    df_names = get_dataframe_names(str_args, [])

    num_strs = 0
    for arg in steps_manager.original_args:
        if isinstance(arg, str):
            df_name = df_names[num_strs]
            num_strs += 1
            # NOTE: we turn this into a raw string with the leading r, so that 
            # Windows \'s aren't read as unicode sequences
            code.append(
                f'{df_name} = pd.read_csv(r\'{arg}\')'
            )

    if len(code) > 0:
        code.insert(0, '# Read in filepaths as dataframes')
            
    return code


PREPROCESS_READ_FILE_PATH = {
    'execute': execute_preprocess_read_file_paths,
    'transpile': transpile_preprocess_read_file_paths
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Utilities to help with type functions
"""

from typing import Tuple, Union
import pandas as pd
import numpy as np

# Because type(1) = int, thus 1 is a 'number' in the Mito type system
MITO_PRIMITIVE_TYPE_MAPPING = {
    'boolean': [bool],
    'timestamp': [pd.Timestamp],
    'timedelta': [pd.Timedelta],
    'number': [int, float],
    'string': [str],
}

BOOLEAN_SERIES = 'boolean_series'
DATETIME_SERIES = 'datetime_series'
TIMEDELTA_SERIES = 'timedelta_series'
NUMBER_SERIES = 'number_series'
STRING_SERIES = 'string_series'

# A series of helper functions that help you figure
# out which dtype we're dealing with. NOTE: since some
# of these types can be different varieties (e.g. int can be int64, uint64)
# we try to check for them with simple expressions

def is_bool_dtype(dtype):
    return 'bool' == dtype

def is_int_dtype(dtype):
    return 'int' in dtype

def is_float_dtype(dtype):
    return 'float' in dtype

def is_string_dtype(dtype):
    return dtype == 'object' or dtype == 'str' or dtype == 'string'

def is_datetime_dtype(dtype):
    # NOTE: this should handle all different datetime columns, no matter
    # the timezone, as it checks for string inclusion
    return 'datetime' in dtype

def is_timedelta_dtype(dtype):
    return 'timedelta' in dtype

def get_float_columns(df: pd.DataFrame):
    """
    Returns all the float column_headers in the
    passed dataframe
    """
    float_columns = []
    for column_header in df.columns:
        if is_float_dtype(str(df[column_header].dtype)):
            float_columns.append(column_header)
    return float_columns

def get_datetime_columns(df: pd.DataFrame):
    """
    Returns all the datetime column_headers in the
    passed dataframe
    """
    datetime_columns = []
    for column_header in df.columns:
        if is_datetime_dtype(str(df[column_header].dtype)):
            datetime_columns.append(column_header)
    return datetime_columns

def get_timedelta_columns(df: pd.DataFrame):
    """
    Returns all the timedelta column_headers in the
    passed dataframe
    """
    timedelta_columns = []
    for column_header in df.columns:
        if is_timedelta_dtype(str(df[column_header].dtype)):
            timedelta_columns.append(column_header)
    return timedelta_columns

def get_mito_type(obj):

    if isinstance(obj, pd.Series):
        dtype = str(obj.dtype)
        if is_bool_dtype(dtype):
            return BOOLEAN_SERIES
        elif is_int_dtype(dtype) or is_float_dtype(dtype):
            return NUMBER_SERIES
        elif is_string_dtype(dtype):
            return STRING_SERIES
        elif is_datetime_dtype(dtype):
            return DATETIME_SERIES
        elif is_timedelta_dtype(dtype):
            return TIMEDELTA_SERIES
        else:
            # We default to string, when not sure what else
            return STRING_SERIES

    elif isinstance(obj, pd.Timestamp):
        return 'timestamp'
    elif isinstance(obj, pd.Timedelta):
        return 'timedelta'
    else:
        obj_type = type(obj)

        for key, value in MITO_PRIMITIVE_TYPE_MAPPING.items():
            if obj_type in value:
                return key

    return None


def get_nan_indexes(*argv): 
    """
    Given the list of arguments to a function, as *argv, 
    returns a list of row indexes that is True iff one of the series 
    params is NaN at that index. Otherwise the index is False

    This function is called by the filter_nan decorator 
    """

    # we find the max length of the args because we are unsure ahead of time which arg 
    # is a series. we need the max_length to construct the boolean index array
    max_length = -1
    for arg in argv:
        # check the type to make sure an error is not thrown for calling len() on an int
        if isinstance(arg, pd.Series) and len(arg) > max_length:
            max_length = len(arg)

    # if there are no series, then:
    #   1. there are no NaN values
    #   2. all of the args are a length of 1
    if max_length == -1:
        return pd.Series([False], dtype='bool')

    nan_indexes = pd.Series([False for i in range(max_length)], dtype='bool')

    # for each row, check for NaN values in the function
    for arg in argv:
        if isinstance(arg, pd.Series): 
            nan_indexes = nan_indexes | pd.isna(arg)

    return nan_indexes

def put_nan_indexes_back(series, nan_indexes):
    """
    This functions takes a series and a boolean index list that is True 
    if the index in the series should be NaN and false if it should
    be left alone. 

    Returns the series with the NaN values put in

    This function is called by the as_types decorator 
    """
    original_length = len(nan_indexes)
    final_series = []
    real_index = 0
    non_nan_index = 0

    while real_index < original_length:
        if nan_indexes[real_index]:
            final_series.append(np.NaN)
        else:
            final_series.append(series[non_nan_index])
            non_nan_index += 1
        real_index +=1

    final_series = pd.Series(final_series)
    return final_series

def get_datetime_format(string_series):
    """
    Given a series of datetime strings, detects if the format is MM-DD-YYYY,
    which is the most common format that pandas does not default to.

    In the future, we can extend this to detect other formats. Returns None
    if infer_datetime_format is good enough!
    """
    try:
        # If we can convert all non null inputs, then we assume that pandas
        # is guessing the input correctly
        non_null_inputs = string_series[~string_series.isna()]
        converted = pd.to_datetime(non_null_inputs, errors='coerce', infer_datetime_format=True)
        if converted.isna().sum() > 0:
            raise Exception("Non full conversion")
        return None
    except:
        # Otherwise, we manually figure out the format.
        sample_string_datetime = string_series[string_series.first_valid_index()]
        if "/" in sample_string_datetime: 
            return '%m/%d/%Y'
        else:
            return '%m-%d-%Y'


from mitosheet.saved_analyses import read_and_upgrade_analysis
from mitosheet.step_performers import STEP_TYPE_TO_STEP_PERFORMER
import json


def get_saved_analysis_description(event, steps_manager):
    """
    Gets the description of the steps in the saved analysis, and sends
    it back to the frontend

    Sends '' if its unable to read in for any reason (or this analysis does)
    not exist.
    """
    analysis_name = event['analysis_name']
    analysis_steps = read_and_upgrade_analysis(analysis_name)
    
    try:
        if analysis_steps is None:
            return ''
        
        analysis_descriptions = []
        for step_data in analysis_steps['steps_data']:
            step_performer = STEP_TYPE_TO_STEP_PERFORMER[step_data['step_type']]

            # Generate the step description
            step_description = step_performer.describe(
                **step_data['params']
            )

            analysis_descriptions.append({
                'step_type': step_data['step_type'],
                'step_description': step_description
            })

        return json.dumps(analysis_descriptions)
        
    except Exception as e:
        # As not being able to get the steps is not a critical failure, 
        # we return empty data if its not possible.
        return ''



from typing import List
import pandas as pd
import plotly.graph_objects as go

from mitosheet.mito_analytics import log
from mitosheet.api.graph.graph_utils import BOX, CREATE_FIG_CODE, SHOW_FIG_CODE, X, filter_df_to_safe_size, get_graph_title, is_all_number_series


def get_box_plot(axis, df, column_headers): 
    """
    Returns a box plot using the following heuristic:

    - If there are 2 or less series, no filtering is required.
    - If there are 3 series, filter to the first BOX_PLOT_3_SERIES_MAX_NUMBER_OF_ROWS rows.
    - If there are 4 series, filter to the first BOX_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS rows.
    """
    fig = go.Figure()

    # Make sure all of the series are NUMBER_SERIES
    if not is_all_number_series(df, column_headers):
        # If not all of the series are numbers, then 
        # return a blank graph
        log(f'failed_generate_graph', {
            'params_graph_type': BOX,
            'params_axis': axis,
            'params_column_headers': column_headers,
            'params_failed_reason': 'non-number-series'
        })
        return go.Figure(), False
      
    df, filtered = filter_df_to_safe_size(BOX, df, column_headers)

    filter_label = ('top 500k') if filtered else ''
    for column_header in column_headers:
        if axis == X:
            fig.add_trace(go.Box(x=df[column_header], name=(' ').join([column_header, filter_label])))
        else:
            fig.add_trace(go.Box(y=df[column_header], name=(' ').join([column_header, filter_label])))

    graph_title = get_graph_title(column_headers, [], filtered, BOX)

    fig.update_layout(
        title = graph_title,
        barmode='stack'
    )

    log(f'generate_graph', {
        'params_graph_type': BOX,
        'params_axis': axis,
        'params_column_headers': column_headers,
        'params_filtered': filtered
    })
    
    return fig, False

def get_box_code(
        axis: str, 
        df: pd.DataFrame, 
        column_headers: List[str],
        df_name: str
    ):
    """
    Generates the code for a box plot, as is generated.
    """
    _, filtered = filter_df_to_safe_size(BOX, df, column_headers)

    filtered_code = ''
    if filtered:
        filtered_code = """
# Filter the dataframe so that it does not crash the browser
from mitosheet import filter_df_to_safe_size
{df_name}_filtered, _ = filter_df_to_safe_size('box', {df_name}, {column_headers})
""".format(
    df_name=df_name, 
    column_headers=column_headers
)       
        df_name = f'{df_name}_filtered'


    graph_title = get_graph_title(column_headers, [], filtered, BOX)

    

    return """{CREATE_FIG_CODE}
{filtered_code}
# Add box plots to the graph
for column_header in {column_headers}:
    fig.add_trace(go.Box({axis}={df_name}[column_header], name=column_header))

# Update the title and stacking mode of the graph
# See Plotly documentation for customizations: https://plotly.com/python/reference/box/
fig.update_layout(
    title='{graph_title}',
    barmode='stack'
)
{SHOW_FIG_CODE}""".format(
    CREATE_FIG_CODE=CREATE_FIG_CODE,
    df_name=df_name, 
    filtered_code=filtered_code,
    column_headers=column_headers, 
    graph_title=graph_title,
    axis=axis.lower(),
    SHOW_FIG_CODE=SHOW_FIG_CODE
)


from enum import Enum
from typing import List
import pandas as pd


"""
    The DataTypeInMito enum is used to signify the type of data in Mito.
    It can be one of four types:

    1. NONE = no data is in Mito
    2. PROVIDED = any data that we provided through our docs, except the tutorial data which has a special designation
    3. TUTORIAL = the tutorial data provided through the docs
    4. PERSONAL = any data not provided by Mito

    NOTE: this should be the same as the enum in Mito.tsx with the same name
"""
class DataTypeInMito(Enum):
    NONE = 'none'
    PROVIDED = 'provided'
    TUTORIAL = 'tutorial'
    PERSONAL = 'personal'


# We keep track of the tutorial and provided data
TUTORIAL_FILE_NAMES = ['Airport-Pets.csv', 'Zipcode-Data.csv']
TUTORIAL_AIRPORT_PETS_COLUMNS = ['Zip', 'City', 'State', 'Division', 'Parking', 'Pets', 'Food', 'Lounge']
TUTORIAL_ZIPCODE_COLUMNS = ['Zip', 'Median_Income', 'Mean_Income', 'Pop']

PROVIDED_TICKET_OFFICE_COLUMNS = ['Zip', 'City', 'State', 'Ticket_Office']
PROVIDED_TICKET_ZIPCODE_COLUMNS = ['Zip', 'Median_Income', 'Median_Income', 'Mean_Income', 'Pop']


def get_data_type_in_mito(dfs: List[pd.DataFrame]) -> DataTypeInMito:
    """
    Returns the DataTypeInMito based on the dataframes passed
    to the function.
    """
    if len(dfs) == 0:
        return DataTypeInMito.NONE

    for df in dfs:
        
        # If the user passed a dataframe with headers the same as the tutorial data
        if (
            df.columns.tolist() == TUTORIAL_AIRPORT_PETS_COLUMNS or 
            df.columns.tolist() == TUTORIAL_ZIPCODE_COLUMNS 
        ): 
            return DataTypeInMito.TUTORIAL

        # If the user passed a dataframe with headers the same as the provided data
        if (
            df.columns.tolist() == PROVIDED_TICKET_OFFICE_COLUMNS or 
            df.columns.tolist() == PROVIDED_TICKET_ZIPCODE_COLUMNS
        ):
            return DataTypeInMito.PROVIDED

    # Otherwise, we have personal data in the tool
    return DataTypeInMito.PERSONAL



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
This preprocessing step is responsible for making a copy of all of the
passed arguments, so that dataframes aren't modified incorrectly.
"""
import pandas as pd
from copy import deepcopy

from mitosheet.mito_analytics import log

def execute_copy(args):
    new_args = []
    for arg in args:
        if isinstance(arg, pd.DataFrame):
            # Do a pandas copy if it's a dataframe
            arg_copy = arg.copy(deep=True)
        else:
            # Simple deepcopy if it's a string
            arg_copy = deepcopy(arg)
        new_args.append(arg_copy)
    return new_args




def transpile_copy(
        steps_manager
    ):
    """
    This always returns nothing; we are just making copies
    """
    return []

PREPROCESS_COPY = {
    'execute': execute_copy,
    'transpile': transpile_copy
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
See mito/mitosheet/steps/README.md for more information about 
how to add a step!
"""

from typing import Dict, List
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.pivot import PivotStepPerformer
from mitosheet.step_performers.filter import FilterStepPerformer
from mitosheet.step_performers.sort import SortStepPerformer
from mitosheet.step_performers.column_steps.reorder_column import ReorderColumnStepPerformer
from mitosheet.step_performers.column_steps.add_column import AddColumnStepPerformer
from mitosheet.step_performers.column_steps.set_column_formula import SetColumnFormulaStepPerformer
from mitosheet.step_performers.merge import MergeStepPerformer
from mitosheet.step_performers.column_steps.delete_column import DeleteColumnStepPerformer
from mitosheet.step_performers.column_steps.rename_column import RenameColumnStepPerformer
from mitosheet.step_performers.column_steps.change_column_dtype import ChangeColumnDtypeStepPerformer
from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer
from mitosheet.step_performers.dataframe_steps.dataframe_delete import DataframeDeleteStepPerformer
from mitosheet.step_performers.dataframe_steps.dataframe_duplicate import DataframeDuplicateStepPerformer
from mitosheet.step_performers.dataframe_steps.dataframe_rename import DataframeRenameStepPerformer

# All steps must be listed in this variable.
STEP_PERFORMERS: List[StepPerformer] = [
    PivotStepPerformer,
    ReorderColumnStepPerformer,
    FilterStepPerformer,
    SortStepPerformer,
    AddColumnStepPerformer,
    SetColumnFormulaStepPerformer,
    ChangeColumnDtypeStepPerformer,
    MergeStepPerformer,
    DeleteColumnStepPerformer,
    RenameColumnStepPerformer,
    SimpleImportStepPerformer,
    DataframeDeleteStepPerformer,
    DataframeDuplicateStepPerformer,
    DataframeRenameStepPerformer,
]

# A helpful mapping for looking up steps based on the incoming events
EVENT_TYPE_TO_STEP_PERFORMER: Dict[str, StepPerformer] = {
    step_performer.step_event_type(): step_performer
    for step_performer in STEP_PERFORMERS
}

# We also build a useful lookup mapping for the step type to step object
STEP_TYPE_TO_STEP_PERFORMER: Dict[str, StepPerformer] = {
    step_performer.step_type(): step_performer
    for step_performer in STEP_PERFORMERS
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import deepcopy
from mitosheet.step_performers.step_performer import StepPerformer
from typing import Any, Dict, List, Optional
from mitosheet.state import State
from mitosheet.parser import safe_replace
from mitosheet.utils import is_valid_header
from mitosheet.step_performers.column_steps.set_column_formula import _update_column_formula_in_step

from mitosheet.errors import (
    make_invalid_column_headers_error,
    make_column_exists_error, make_no_sheet_error,
)

class RenameColumnStepPerformer(StepPerformer):
    """"
    A rename_column step, which allows you to rename a column
    in a dataframe.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'rename_column' 

    @classmethod
    def step_display_name(cls) -> str:
        return 'Renamed a Column'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'rename_column_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        old_column_header: str,
        new_column_header: str,
        **params
    ) -> State:
        # if the sheet doesn't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error(sheet_index)

        if not is_valid_header(new_column_header):
            raise make_invalid_column_headers_error([new_column_header])

        if new_column_header in prev_state.dfs[sheet_index].keys():
            raise make_column_exists_error(new_column_header)

        # Create a new post state for this step
        post_state = deepcopy(prev_state)

        # Execute the rename
        post_state.dfs[sheet_index].rename(columns={old_column_header: new_column_header}, inplace=True)

        # Then, we update the current step to be valid, namely by deleting the old column (wherever it is)
        # and replacing it with the new column. 
        sheet_column_metatype = post_state.column_metatype[sheet_index]
        sheet_column_metatype[new_column_header] = sheet_column_metatype[old_column_header]

        sheet_column_type = post_state.column_type[sheet_index]
        sheet_column_type[new_column_header] = sheet_column_type[old_column_header]

        sheet_column_spreadsheet_code = post_state.column_spreadsheet_code[sheet_index]
        sheet_column_spreadsheet_code[new_column_header] = sheet_column_spreadsheet_code[old_column_header]

        sheet_column_python_code = post_state.column_python_code[sheet_index]
        sheet_column_python_code[new_column_header] = sheet_column_python_code[old_column_header].replace(
            f'df[\'{old_column_header}\']',
            f'df[\'{new_column_header}\']'
        )
        
        sheet_column_evaluation_graph = post_state.column_evaluation_graph[sheet_index]
        sheet_column_evaluation_graph[new_column_header] = sheet_column_evaluation_graph[old_column_header]

        sheet_column_filters = post_state.column_filters[sheet_index]
        sheet_column_filters[new_column_header] = sheet_column_filters[old_column_header]

        # We also have to go over _all_ the formulas in the sheet that reference this column, and update
        # their references to the new column. 
        for column_header in sheet_column_evaluation_graph[new_column_header]:
            old_formula = sheet_column_spreadsheet_code[column_header]
            new_formula = safe_replace(
                old_formula,
                old_column_header,
                new_column_header
            )

            # NOTE: this does not update the evaluation graph for columns that are descendents
            # of this column, so we do that below.
            _update_column_formula_in_step(post_state, sheet_index, column_header, old_formula, new_formula)

        # We then have to go through and update the evaluation graphs
        # for the columns the renamed column relied on.
        for dependents in sheet_column_evaluation_graph.values():
            if old_column_header in dependents:
                dependents.remove(old_column_header)
                dependents.add(new_column_header)

        # We delete all references to the old_column header
        # NOTE: this has to happen after the above formula setting, so that
        # the dependencies can be updated properly!
        del sheet_column_metatype[old_column_header]
        del sheet_column_type[old_column_header]
        del sheet_column_spreadsheet_code[old_column_header]
        del sheet_column_python_code[old_column_header]
        del sheet_column_evaluation_graph[old_column_header]
        del sheet_column_filters[old_column_header]

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        old_column_header: str,
        new_column_header: str
    ) -> List[str]:
        df_name = post_state.df_names[sheet_index]
        rename_dict = "{\"" + old_column_header + "\": \"" + new_column_header + "\"}"
        return [f'{df_name}.rename(columns={rename_dict}, inplace=True)']

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        old_column_header: str,
        new_column_header: str,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_name = df_names[sheet_index]
            return f'Renamed {old_column_header} to {new_column_header} in {df_name}'
        return f'Renamed {old_column_header} to {new_column_header}'


"""
Contains tests for the user.json file, making sure it upgrades properly,
and that it can handle the various undefined versions of user.json that exist.
"""
import os
from copy import deepcopy
import subprocess

from mitosheet.utils import get_random_id
from mitosheet._version import __version__
from mitosheet.user.schemas import UJ_MITOSHEET_TELEMETRY, USER_JSON_VERSION_1, USER_JSON_VERSION_2
from mitosheet.user.db import USER_JSON_PATH, get_user_field
from mitosheet.user import initialize_user
from mitosheet.tests.user.conftest import check_user_json, write_fake_user_json, cleanup_files, today_str
from mitosheet.user.schemas import (
    UJ_INTENDED_BEHAVIOR, UJ_CLOSED_FEEDBACK, UJ_MITOSHEET_LAST_FIVE_USAGES,
    UJ_USER_JSON_VERSION, UJ_STATIC_USER_ID, UJ_USER_SALT, UJ_USER_EMAIL, 
    UJ_RECEIVED_TOURS, UJ_FEEDBACKS, UJ_MITOSHEET_CURRENT_VERSION, 
    UJ_MITOSHEET_LAST_UPGRADED_DATE
)


def test_user_json_version_one_is_final():
    # NOTE: this final version should never change, as it is final!
    assert set(USER_JSON_VERSION_1.keys()) == \
        set([
            UJ_USER_JSON_VERSION, 
            UJ_STATIC_USER_ID, 
            UJ_USER_SALT, 
            UJ_USER_EMAIL, 
            UJ_INTENDED_BEHAVIOR, 
            UJ_RECEIVED_TOURS, 
            UJ_FEEDBACKS, 
            UJ_CLOSED_FEEDBACK, 
            UJ_MITOSHEET_CURRENT_VERSION, 
            UJ_MITOSHEET_LAST_UPGRADED_DATE, 
            UJ_MITOSHEET_LAST_FIVE_USAGES
        ])


def test_initialize_user_creates_json_when_empty():
    initialize_user()
    check_user_json()
    os.remove(USER_JSON_PATH)


def test_initialize_does_not_recreate_json_when_exists():
    initialize_user()
    user_id_old = get_user_field(UJ_STATIC_USER_ID)
    check_user_json()

    initialize_user()
    user_id_new = get_user_field(UJ_STATIC_USER_ID)
    check_user_json()

    assert user_id_new == user_id_old
    os.remove(USER_JSON_PATH)


def test_initialize_upgrades_user_when_oldest_version():
    write_fake_user_json({
        UJ_STATIC_USER_ID: 'github_action',
        UJ_USER_EMAIL: 'github@action.com'
    })
    initialize_user()
    check_user_json(user_email='github@action.com')

    os.remove(USER_JSON_PATH)


def test_initialize_removed_unused_feilds():
    write_fake_user_json({
        UJ_STATIC_USER_ID: 'github_action',
    })
    initialize_user()
    assert get_user_field(UJ_MITOSHEET_LAST_FIVE_USAGES) == None

def test_initialize_upgrades_user_when_installer_version():
    write_fake_user_json({
        UJ_STATIC_USER_ID: 'github_action',
    })
    initialize_user()
    check_user_json(user_email='')

    os.remove(USER_JSON_PATH)


def test_initialize_upgrades_missing_any_key():

    for key in USER_JSON_VERSION_1:
        user_json_missing_key = deepcopy(USER_JSON_VERSION_1)
        del user_json_missing_key[key]
        write_fake_user_json(user_json_missing_key, static_user_id=get_random_id())
        initialize_user()
        check_user_json(user_email='')

    os.remove(USER_JSON_PATH)


def test_initialize_upgrades_missing_multiple_keys():

    key_sets = [
        [UJ_INTENDED_BEHAVIOR, UJ_RECEIVED_TOURS, UJ_FEEDBACKS, UJ_CLOSED_FEEDBACK],
        [UJ_USER_JSON_VERSION, UJ_USER_SALT, UJ_CLOSED_FEEDBACK],
        [UJ_USER_JSON_VERSION, UJ_INTENDED_BEHAVIOR, UJ_RECEIVED_TOURS, UJ_FEEDBACKS, UJ_CLOSED_FEEDBACK, UJ_USER_SALT],
    ]

    for key_set in key_sets:
        user_json_missing_keys = deepcopy(USER_JSON_VERSION_1)
        for key in key_set:
            del user_json_missing_keys[key]

        write_fake_user_json(user_json_missing_keys, static_user_id=get_random_id())
        initialize_user()
        check_user_json(user_email='')

    os.remove(USER_JSON_PATH)


def test_initalize_missing_static_user_id():

    user_json_missing_static_user_id = deepcopy(USER_JSON_VERSION_1)
    del user_json_missing_static_user_id[UJ_STATIC_USER_ID]
    
    write_fake_user_json(user_json_missing_static_user_id)
    initialize_user()
    check_user_json(user_email='')

    os.remove(USER_JSON_PATH)


def test_initalize_creates_valid_user_json_from_invalid():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write("Hi, 123")
    
    initialize_user()
    check_user_json()


def test_initalize_creates_valid_user_json_with_failed_upgrade():
    initialize_user()

    # TODO: figure out how to cause an error in the middle of an upgrade
    pass

def test_initalize_upgrades_and_adds_new_usages():

    write_fake_user_json(
        USER_JSON_VERSION_1, 
        static_user_id=get_random_id(),
        mitosheet_last_five_usages=['2020-12-1']
    )

    initialize_user()
    check_user_json(
        mitosheet_last_fifty_usages=['2020-12-1', today_str],
        user_email=''
    )

    os.remove(USER_JSON_PATH)



def test_initalize_detects_new_usage_when_over_fifty():
    write_fake_user_json(
        USER_JSON_VERSION_2,
        mitosheet_last_fifty_usages=['2020-12-1'] * 50
    )

    initialize_user()
    check_user_json(
        mitosheet_last_fifty_usages=['2020-12-1'] * 49 + [today_str],
        user_email=''
    )

    os.remove(USER_JSON_PATH)


def test_initalize_only_stores_each_day_once():
    write_fake_user_json(
        USER_JSON_VERSION_2,
        mitosheet_last_fifty_usages=['2020-12-1'] * 49 + [today_str]
    )

    initialize_user()
    check_user_json(
        mitosheet_last_fifty_usages=['2020-12-1'] * 49 + [today_str],
        user_email=''
    )

    os.remove(USER_JSON_PATH)

    
def test_initalize_detects_when_mito_upgraded():
    write_fake_user_json(
        USER_JSON_VERSION_2,
        mitosheet_current_version='0.1.100',
        mitosheet_last_upgraded_date='2020-12-1',
    )

    initialize_user()
    check_user_json(
        mitosheet_current_version=__version__,
        mitosheet_last_upgraded_date=today_str,
        user_email=''
    )

    os.remove(USER_JSON_PATH)


def test_initialize_doesnt_overwrite_feedback():
    write_fake_user_json(
        USER_JSON_VERSION_1,
        static_user_id=get_random_id(),
        feedbacks=['A', 'B']
    )

    initialize_user()
    check_user_json(
        feedbacks=['A', 'B'],
        user_email=''
    )

def test_upgrades_version_2_to_version_3():
    write_fake_user_json(
        USER_JSON_VERSION_2,
        static_user_id=get_random_id()
    )

    initialize_user()

    check_user_json(
        user_email=''
    )

def test_main_cli_turn_off_logging():
    initialize_user()
    out = subprocess.run(['python', '-m', 'mitosheet', 'turnofflogging'], capture_output=True)
    assert get_user_field(UJ_MITOSHEET_TELEMETRY) == False


"""
Helpers for accessing the user.json file
"""
import os
import json

# Where all global .mito files are stored
MITO_FOLDER = os.path.join(os.path.expanduser("~"), '.mito')

# The path of the user.json file
USER_JSON_PATH = os.path.join(MITO_FOLDER, 'user.json')

def get_user_json_object():
    """
    Gets the entire user json object
    """
    try:
        with open(USER_JSON_PATH) as f:
            return json.load(f)
    except: 
        return None

def get_user_field(field):
    """
    Returns the value stored at field in the user.json file,
    but may read a different file if it passed
    """
    try:
        with open(USER_JSON_PATH) as f:
            return json.load(f)[field]
    except: 
        return None

def set_user_json_object(user_json_object):
    """
    Updates the value of a specific feild in user.json
    """
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps(user_json_object))

def set_user_field(field, value):
    """
    Updates the value of a specific feild in user.json
    """
    with open(USER_JSON_PATH, 'r') as user_file_old:
        old_user_json = json.load(user_file_old)
        old_user_json[field] = value
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(old_user_json))



from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer
from mitosheet.saved_analyses import read_and_upgrade_analysis

def get_import_summary(event):
    """
    Handle import summary is a route that, given the name of an analysis, will
    return the parameters to import steps over the course of the analysis. 

    The data we return is in the format:
    {
        "1": ["file123.csv"], 
        "3": ["file12356.csv"]
    }
    which is a mapping from _step index_ to the files that they import.
    """
    analysis_name = event['analysis_name']
    analysis = read_and_upgrade_analysis(analysis_name)

    imports_only = dict()
    if analysis is not None:
        for step_idx, step_data in enumerate(analysis['steps_data']):
            if step_data['step_type'] == SimpleImportStepPerformer.step_type():
                # NOTE: we make the step indexes strings because JSON insists that
                # all keys are strings anyways
                imports_only[str(step_idx)] = step_data['params']['file_names']

    return imports_only


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import deepcopy
from typing import List, OrderedDict
import pandas as pd

from mitosheet.sheet_functions.types.utils import get_mito_type

# Constants for where the dataframe in the state came from
DATAFRAME_SOURCE_PASSED = 'passed' # passed in mitosheet.sheet
DATAFRAME_SOURCE_IMPORTED = 'imported' # imported through a simple import
DATAFRAME_SOURCE_PIVOTED = 'pivoted' # created through a pivot
DATAFRAME_SOURCE_MERGED = 'merged' # created through a merge
DATAFRAME_SOURCE_DUPLICATED = 'duplicated' # created through a sheet duplication


class State():
    """
    State is a container that stores the current state of a Mito analysis,
    where each step that is applied takes a state as input and creates a
    new state as output. 

    It stores the obvious things, like the dataframes and their names, but 
    also other helper pieces of state like: the column formulas, the filters,
    etc.
    """

    def __init__(self, 
            dfs: List[pd.DataFrame], 
            df_names: List[str]=None,
            df_sources: List[str]=None,
            column_metatype=None,
            column_type=None,
            column_spreadsheet_code=None,
            column_python_code=None,
            column_evaluation_graph=None,
            column_filters=None
        ):

        # The dataframes that are in the state
        self.dfs = dfs

        # The df_names are composed of two parts:
        # 1. The names of the variables passed into the mitosheet.sheet call (which don't change over time).
        # 2. The names of the dataframes that were created during the analysis (e.g. by a merge).
        # Until we get them from the frontend as an update_event, we default them to df1, df2, ...
        self.df_names = df_names if df_names is not None else [f'df{i + 1}' for i in range(len(dfs))] 

        # The df sources are where the actual dataframes come from, e.g.
        # how the dataframe was created. If not df sources passed, then this is in the
        # initialize state, and so these dataframes were passed to the mitosheet
        # call
        self.df_sources = df_sources if df_sources is not None else [DATAFRAME_SOURCE_PASSED for _ in dfs]

        # The column_metatype is if it stores formulas or values
        self.column_metatype = column_metatype if column_metatype is not None else [{key: 'value' for key in df.keys()} for df in dfs]

        # The column_type is the type of the series in this column 
        self.column_type = column_type if column_type is not None else [{key: get_mito_type(df[key]) for key in df.keys()} for df in dfs]

        # We make column_spreadsheet_code an ordered dictonary to preserve the order the formulas
        # are inserted, which in turn makes sure when we save + rerun an analysis, it's recreated
        # in the correct order (and thus the column order is preserved).
        self.column_spreadsheet_code = column_spreadsheet_code if column_spreadsheet_code is not None else [OrderedDict({key: '' for key in df.keys()}) for df in dfs]
        self.column_python_code = column_python_code if column_python_code is not None else [{key: '' for key in df.keys()} for df in dfs]
        self.column_evaluation_graph = column_evaluation_graph if column_evaluation_graph is not None else [{key: set() for key in df.keys()} for df in dfs]
        self.column_filters = column_filters if column_filters is not None else [{key: {'operator': 'And', 'filters': []} for key in df.keys()} for df in dfs]

    
    def __copy__(self):
        """
        If you copy a state using the copy() function, this Python
        function is called, and returns a shallow copy of the state
        """
        return State(
            [df.copy(deep=False) for df in self.dfs],
            df_names=deepcopy(self.df_names),
            df_sources=deepcopy(self.df_sources),
            column_metatype=deepcopy(self.column_metatype),
            column_type=deepcopy(self.column_type),
            column_spreadsheet_code=deepcopy(self.column_spreadsheet_code),
            column_python_code=deepcopy(self.column_python_code),
            column_evaluation_graph=deepcopy(self.column_evaluation_graph),
            column_filters=deepcopy(self.column_filters)
        )


    def __deepcopy__(self, memo):
        """
        If you copy a state using the deepcopy() function, this Python
        function is called, and returns a deep copy of the state
        """
        return State(
            [df.copy(deep=True) for df in self.dfs],
            df_names=deepcopy(self.df_names),
            df_sources=deepcopy(self.df_sources),
            column_metatype=deepcopy(self.column_metatype),
            column_type=deepcopy(self.column_type),
            column_spreadsheet_code=deepcopy(self.column_spreadsheet_code),
            column_python_code=deepcopy(self.column_python_code),
            column_evaluation_graph=deepcopy(self.column_evaluation_graph),
            column_filters=deepcopy(self.column_filters)
        )

    def add_df_to_state(
            self, 
            new_df: pd.DataFrame, 
            df_source: str, 
            sheet_index=None,
            df_name=None
        ):
        """
        Helper function for adding a new dataframe to this state,
        and keeping all the other variables in sync.

        If sheet_index is defined, then will replace the dataframe
        that is currently at the index. Otherwise, if sheet_index is
        not defined, then will append the df to the end of the state
        """
        if sheet_index is None:
            # Update dfs by appending new df
            self.dfs.append(new_df)
            # Also update the dataframe name
            if df_name is None:
                self.df_names.append(f'df{len(self.df_names) + 1}')
            else:
                self.df_names.append(df_name)

            # Save the source of this dataframe
            self.df_sources.append(df_source)
            
            # Update all the variables that depend on column_headers
            column_headers = new_df.keys()
            self.column_metatype.append({column_header: 'value' for column_header in column_headers})
            self.column_type.append({column_header: get_mito_type(new_df[column_header]) for column_header in column_headers})
            self.column_spreadsheet_code.append({column_header: '' for column_header in column_headers})
            self.column_python_code.append({column_header: '' for column_header in column_headers})
            self.column_evaluation_graph.append({column_header: set() for column_header in column_headers})
            self.column_filters.append({column_header: {'operator':'And', 'filters': []} for column_header in column_headers})

            # Return the index of this sheet
            return len(self.dfs) - 1
        else:

            # Update dfs by switching which df is at this index specifically
            self.dfs[sheet_index] = new_df
            # Also update the dataframe name, if it is passed. Otherwise, we don't change it
            if df_name is not None:
                self.df_names[sheet_index] = df_name

            # Save the source of this dataframe, if it is passed. Otherwise, don't change it
            if df_source is not None:
                self.df_sources[sheet_index] = df_source
            
            # Update all the variables that depend on column_headers
            column_headers = new_df.keys()
            self.column_metatype[sheet_index] = {column_header: 'value' for column_header in column_headers}
            self.column_type[sheet_index] = {column_header: get_mito_type(new_df[column_header]) for column_header in column_headers}
            self.column_spreadsheet_code[sheet_index] = {column_header: '' for column_header in column_headers}
            self.column_python_code[sheet_index] = {column_header: '' for column_header in column_headers}
            self.column_evaluation_graph[sheet_index] = {column_header: set() for column_header in column_headers}
            self.column_filters[sheet_index] = {column_header: {'operator':'And', 'filters': []} for column_header in column_headers}

            # Return the index of this sheet
            return sheet_index
    
    def does_sheet_index_exist_within_state(self, sheet_index):
        """
        Returns true iff a sheet_index exists within this state
        """
        return not (sheet_index < 0 or sheet_index >= len(self.dfs))

    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pytest
import pandas as pd

from mitosheet.transpiler.transpile import transpile
from mitosheet.tests.test_utils import create_mito_wrapper, create_mito_wrapper_dfs


def test_transpile_single_column():
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)

    assert mito.transpiled_code == [
        "df1.insert(1, 'B', 0)", 
        'df1[\'B\'] = df1[\'A\']'
    ]


def test_transpile_multiple_columns_no_relationship():
    mito = create_mito_wrapper(['abc'])
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')
    code_container = transpile(mito.mito_widget.steps_manager)
    
    assert len([line for line in code_container['code'] if 'df1.insert(1, \'B\', 0)' in line]) > 0
    assert len([line for line in code_container['code'] if 'df1.insert(2, \'C\', 0)' in line]) > 0

def test_transpile_columns_in_each_sheet():
    mito = create_mito_wrapper(['abc'], sheet_two_A_data=['abc'])
    mito.add_column(0, 'B')
    mito.add_column(1, 'B')
    code_container = transpile(mito.mito_widget.steps_manager)
    assert len([line for line in code_container['code'] if 'df1.insert(1, \'B\', 0)' in line]) > 0
    assert len([line for line in code_container['code'] if 'df2.insert(1, \'B\', 0)' in line]) > 0

def test_transpile_multiple_columns_linear():
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)
    mito.set_formula('=B', 0, 'C', add_column=True)

    assert mito.transpiled_code == [
        'df1.insert(1, \'B\', 0)',
        'df1[\'B\'] = df1[\'A\']', 
        'df1.insert(2, \'C\', 0)',
        'df1[\'C\'] = df1[\'B\']'
    ]

COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet(column_header):
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)

    assert mito.transpiled_code == [
        f'df1.insert(1, \'{column_header}\', 0)', 
        f'df1[\'{column_header}\'] = df1[\'A\']'
    ]


COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet_multi_sheet(column_header):
    mito = create_mito_wrapper(['abc'], sheet_two_A_data=['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)
    mito.set_formula('=A', 1, column_header, add_column=True)

    assert mito.transpiled_code == [
        f'df1.insert(1, \'{column_header}\', 0)', 
        f'df1[\'{column_header}\'] = df1[\'A\']', 
        f'df2.insert(1, \'{column_header}\', 0)', 
        f'df2[\'{column_header}\'] = df2[\'A\']'
    ]

def test_preserves_order_columns():
    mito = create_mito_wrapper(['abc'])
    # Topological sort will currently display this in C, B order
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')

    assert mito.transpiled_code == [
        'df1.insert(1, \'B\', 0)',
        'df1.insert(2, \'C\', 0)',
    ]

def test_transpile_delete_column():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1)
    mito.delete_column(0, 'C')

    assert mito.transpiled_code == [
        'df1.drop(\'C\', axis=1, inplace=True)'
    ]


# TESTING OPTIMIZATION

def test_removes_unedited_formulas_for_unedited_sheets():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)

    mito.set_formula('=C', 0, 'D', add_column=True)
    mito.set_formula('=C', 1, 'D', add_column=True)

    mito.merge_sheets('lookup', 0, 'A', ['A', 'B', 'C', 'D'], 1, 'A', ['A', 'B', 'C', 'D'])

    mito.set_formula('=C + 1', 1, 'D', add_column=True)

    assert mito.transpiled_code == [
        "df1.insert(3, 'D', 0)", 
        'df1[\'D\'] = df1[\'C\']',
        "df2.insert(3, 'D', 0)", 
        'df2[\'D\'] = df2[\'C\']',
        'temp_df = df2.drop_duplicates(subset=\'A\') # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'df2[\'D\'] = df2[\'C\'] + 1',
    ]


def test_mulitple_merges_no_formula_steps():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.merge_sheets('lookup', 0, 'A', ['A', 'B', 'C'], 1, 'A', ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 'A', ['A', 'B', 'C'], 1, 'A', ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 'A', ['A', 'B', 'C'], 1, 'A', ['A', 'B', 'C'])


    assert mito.transpiled_code == [
        'temp_df = df2.drop_duplicates(subset=\'A\') # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'temp_df = df2.drop_duplicates(subset=\'A\') # Remove duplicates so lookup merge only returns first match', 
        'df4 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'temp_df = df2.drop_duplicates(subset=\'A\') # Remove duplicates so lookup merge only returns first match', 
        'df5 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])'
    ]

def test_optimization_with_other_edits():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.add_column(0, 'D')
    mito.set_formula('=A', 0, 'D')
    mito.merge_sheets('lookup', 0, 'A', ['A', 'B', 'C', 'D'], 1, 'A', ['A', 'B', 'C'])
    mito.add_column(0, 'AAA')
    mito.delete_column(0, 'AAA')

    assert mito.transpiled_code == [
        "df1.insert(3, 'D', 0)", 
        'df1[\'D\'] = df1[\'A\']',
        'temp_df = df2.drop_duplicates(subset=\'A\') # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'df1.insert(4, \'AAA\', 0)',
        'df1.drop(\'AAA\', axis=1, inplace=True)'
    ]

TAB = '    '
NEWLINE_TAB = f'\n{TAB}'
flatten_code = f'pivot_table.columns = [{NEWLINE_TAB}\'_\'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col{NEWLINE_TAB}for col in pivot_table.columns.values\n]'


def test_transpile_initial_rename_mulitple_columns():
    df1 = pd.DataFrame(data={'First Name': ['Nate', 'Nate'], 'Last Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1)

    assert mito.transpiled_code == [
        '# Rename headers to make them work with Mito',
        'df1.rename(columns={\"First Name\": \"First_Name\", \"Last Name\": \"Last_Name\"}, inplace=True)',
    ]

def test_transpile_initial_rename_multiple_dfs():
    df1 = pd.DataFrame(data={'First Name': ['Nate', 'Nate'], 'Last_Name': ['Rush', 'Jack']})
    df2 = pd.DataFrame(data={'First_Name': ['Nate', 'Nate'], 'Last Name': ['Rush', 'Jack']})
    mito = create_mito_wrapper_dfs(df1, df2)

    assert mito.transpiled_code == [
        '# Rename headers to make them work with Mito',
        'df1.rename(columns={\"First Name\": \"First_Name\"}, inplace=True)',
        'df2.rename(columns={\"Last Name\": \"Last_Name\"}, inplace=True)',
    ]
    
def test_transpile_reorder_column():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 1)

    assert mito.transpiled_code == [
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]'
    ]

def test_transpile_two_column_reorders():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 1)
    mito.reorder_column(0, 'B', 1)

    assert mito.transpiled_code == [
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        'df1_columns = [col for col in df1.columns if col != \'B\']',
        'df1_columns.insert(1, \'B\')',
        'df1 = df1[df1_columns]'
    ]

def test_transpile_reorder_column_invalid():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 5)

    assert mito.transpiled_code == [
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
    ]

def test_transpile_sort_ascending_valid():
    df1 = pd.DataFrame(data={'A': [1, 2, 3, 4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.sort(0, 'A', 'ascending')

    assert mito.transpiled_code == [
        'df1 = df1.sort_values(by=\'A\', ascending=True, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_sort_descending_valid():
    df1 = pd.DataFrame(data={'A': [1, 2, 3, 4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.sort(0, 'A', 'descending')

    assert mito.transpiled_code == [
        'df1 = df1.sort_values(by=\'A\', ascending=False, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)',
    ]

def test_transpile_sort_ascending_then_descending_valid():
    df1 = pd.DataFrame(data={'A': [1, 2, 3, 4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.sort(0, 'A', 'ascending')
    mito.sort(0, 'A', 'descending')

    assert mito.transpiled_code == [
        'df1 = df1.sort_values(by=\'A\', ascending=True, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)',
        'df1 = df1.sort_values(by=\'A\', ascending=False, na_position=\'first\')',
        'df1 = df1.reset_index(drop=True)'
    ]

def test_transpile_merge_then_sort():
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df2 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Sign': ['Gemini', "Tarus"]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.merge_sheets('lookup', 0, 'Name', list(df1.keys()), 1, 'Name', list(df2.keys()))
    mito.sort(2, 'Number', 'ascending')

    assert mito.transpiled_code == [
        'temp_df = df2.drop_duplicates(subset=\'Name\') # Remove duplicates so lookup merge only returns first match',
        'df3 = df1.merge(temp_df, left_on=[\'Name\'], right_on=[\'Name\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'df3 = df3.sort_values(by=\'Number\', ascending=True, na_position=\'first\')',
        'df3 = df3.reset_index(drop=True)'
    ]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
For going to a datetime series.
"""

from typing import Tuple, Union
import pandas as pd
import numpy as np

from mitosheet.sheet_functions.types.utils import BOOLEAN_SERIES, DATETIME_SERIES, NUMBER_SERIES, STRING_SERIES, get_mito_type, get_datetime_format

def to_datetime_series(
        unknown_object,
        on_uncastable_arg_element=('default', np.NaN), # Union[Literal['error'], Tuple[Literal['default'], any]]
    ):
    """
    Converts the given object to a datetime series.
    """
    from_type = get_mito_type(unknown_object)

    # If it is not a series, we put it in a series, and get the type again
    if not from_type.endswith('series'):
        unknown_object = pd.Series([unknown_object])
        from_type = get_mito_type(unknown_object)

    if from_type == BOOLEAN_SERIES:
        return None
    elif from_type == DATETIME_SERIES:
        return unknown_object
    if from_type == NUMBER_SERIES:
        return None
    elif from_type == STRING_SERIES:
        # TODO: improve this to work element wise!
        datetime_format = get_datetime_format(unknown_object)
        if datetime_format is not None:
            return pd.to_datetime(
                unknown_object,
                format=datetime_format,
                errors='coerce'
            )
        else:
            return pd.to_datetime(
                unknown_object,
                infer_datetime_format=True,
                errors='coerce'
            )
    else:
        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
This preprocessing step is responsible for making sure the arguments
passed are of the right type.
"""
import pandas as pd

from mitosheet.mito_analytics import log

def execute_check_args_type(args):
    # We first validate all the parameters as either dataframes or strings
    # but we also allow users to pass None values, which we just ignore (this
    # makes variable number of inputs to the sheet possible).
    
    # NOTE: if passing None values to have them ignored, the user should 
    # be careful to place them at the end of the arguments (so df names)
    # read properly and match up with the correct variables
    for arg in args:
        if not isinstance(arg, pd.DataFrame) and not isinstance(arg, str) and not arg is None:
            error_message = f'Invalid argument passed to sheet: {arg}. Please pass all dataframes or paths to CSV files.'
            log('mitosheet_sheet_call_failed', {'error': error_message})
            raise ValueError(error_message)
    
    # We do filter out all the None arguments
    return [arg for arg in args if arg is not None]




def transpile_check_args_type(
        steps_manager
    ):
    """
    This always returns nothing, because there it is just a sanity check
    and does not actually generate any code.
    """
    return []

PREPROCESS_CHECK_ARGS_TYPE = {
    'execute': execute_check_args_type,
    'transpile': transpile_check_args_type
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Helper functions for timing the Mito codebase in the least intrusive
way possible :) 
"""

import time

# Change this if you don't want to print anything
PRINT_TIMING = False

def timeit(function):
    """
    A decorator that makes it easy to time a function by just adding @timeit. 

    The PRINT_TIMING variable aboves controls what is done with the resulting
    calculation (e.g. should it be printed, or what). 

    TODO: factor PRINT_TIMING out into a config file?
    """

    # A decorator for timing a function!
    # TODO: make this take the parameter from the sheet function call that 
    def timed(*args, **kwargs):
        time_start = time.time()
        result = function(*args, **kwargs)
        time_end = time.time()
        if PRINT_TIMING:
            print(f'{function.__name__} took {time_end - time_start} seconds.')

        return result 
    return timed


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import copy


REDO_EVENT = 'redo'
REDO_PARAMS = []

def execute_redo_update(steps_manager):
    steps_manager.execute_redo()

REDO_UPDATE = {
    'event_type': REDO_EVENT,
    'params': REDO_PARAMS,
    'execute': execute_redo_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains all functions that can be used in a sheet that operate on
datetime objects.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
from mitosheet.sheet_functions.types.utils import DATETIME_SERIES
import pandas as pd

from mitosheet.sheet_functions.types.decorators import convert_arg_to_series_type, filter_nans, handle_sheet_function_errors


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    DATETIME_SERIES,
    'error',
    'error'
)
def DATEVALUE(datetime_series) -> pd.Series:
    """
    {
        "function": "DATEVALUE",
        "description": "Converts a given string to a date series.",
        "examples": [
            "DATEVALUE(A)",
            "DATEVALUE('2012-12-22')"
        ],
        "syntax": "DATEVALUE(date_string)",
        "syntax_elements": [{
                "element": "date_string",
                "description": "The date string to turn into a date object."
            }
        ]
    }
    """
    return datetime_series


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    DATETIME_SERIES,
    'error',
    'error'
)
def DAY(datetime_series) -> pd.Series:
    """
    {
        "function": "DAY",
        "description": "Returns the day of the month that a specific date falls on, as a number.",
        "examples": [
            "DAY(A)",
            "DAY('2012-12-22')"
        ],
        "syntax": "DAY(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the day of."
            }
        ]
    }
    """
    return datetime_series.dt.day


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    DATETIME_SERIES,
    'error',
    'error'
)
def MONTH(datetime_series) -> pd.Series:
    """
    {
        "function": "MONTH",
        "description": "Returns the month that a specific date falls in, as a number.",
        "examples": [
            "MONTH(A)",
            "MONTH('2012-12-22')"
        ],
        "syntax": "MONTH(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the month of."
            }
        ]
    }
    """
    return datetime_series.dt.month


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    DATETIME_SERIES,
    'error',
    'error'
)
def WEEKDAY(datetime_series) -> pd.Series:
    """
    {
        "function": "WEEKDAY",
        "description": "Returns the day of the week that a specific date falls on. 1-7 corresponds to Monday-Sunday.",
        "examples": [
            "WEEKDAY(A)",
            "WEEKDAY('2012-12-22')"
        ],
        "syntax": "WEEKDAY(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the weekday of."
            }
        ]
    }
    """
    return datetime_series.dt.weekday + 1


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    DATETIME_SERIES,
    'error',
    'error'
)
def YEAR(datetime_series) -> pd.Series:
    """
    {
        "function": "YEAR",
        "description": "Returns the day of the year that a specific date falls in, as a number.",
        "examples": [
            "YEAR(A)",
            "YEAR('2012-12-22')"
        ],
        "syntax": "YEAR(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the month of."
            }
        ]
    }
    """
    return datetime_series.dt.year


# TODO: we should see if we can list these automatically!
DATE_FUNCTIONS = {
    'DATEVALUE': DATEVALUE,
    'DAY': DAY,
    'MONTH': MONTH,
    'WEEKDAY': WEEKDAY,
    'YEAR': YEAR,
}


"""
Stores utilities for changing the schemas for a saved analysis file,
which requires a seperate upgrade process that upgrading specific
steps within the file. 
"""


def upgrade_saved_analysis_format_to_steps_data(saved_analysis):
    """
    Upgrades a saved analysis to the current format without upgrading
    the specific individual steps! The current format of the saved analysis 
    is:

    {
        "version": "0.1.197",
        "steps_data": [
            {
                "step_version": 1,
                "step_type": "filter",
                "params": {
                    ....
                }
            }
        ]
    }

    The old version of the saved analysis was (note: it was called
    steps, and the params were just inlined in the object):
    {
        "version": "0.1.197",
        "steps": {
            1: {
                "step_version": 1,
                "step_type": "filter",
                "sheet_index": 1,
                ...
            }
        }
    }
    """ 
    # If this saved analysis has the old format, we update it
    if saved_analysis is None:
        return None

    if 'steps' in saved_analysis:
        version = saved_analysis['version']
        steps_data = saved_analysis['steps']

        new_steps_data = []
        for step_data in steps_data.values():
            step_version = step_data['step_version']
            step_type = step_data['step_type']
            params = {
                key: value for key, value in step_data.items() if (key != 'step_version' and key != 'step_type')
            }

            new_steps_data.append({
                'step_version': step_version,
                'step_type': step_type,
                'params': params
            })
        
        return {
            'version': version,
            'steps_data': new_steps_data
        }
    else:
        return saved_analysis


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from mitosheet.step_performers.step_performer import StepPerformer
import os
from typing import Any, Dict, List, Optional
import pandas as pd
from copy import copy
import json
import csv

from mitosheet.mito_analytics import log
from mitosheet.errors import make_is_directory_error
from os.path import basename, normpath
from mitosheet.utils import get_header_renames, make_valid_header
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State


class SimpleImportStepPerformer(StepPerformer):
    """
    A simple import, which allows you to import dataframes 
    with the given file_names, while detecting the correct
    way to import them.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'simple_import'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Imported CSV Files'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'simple_import_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        file_names: List[str],
        **params
    ) -> State:
        # If any of the files are directories, we throw an error to let
        # the user know
        for file_name in file_names:
            if os.path.isdir(file_name):
                raise make_is_directory_error(file_name)

        # Create a new step
        post_state = copy(prev_state)

        column_header_renames = dict()
        file_delimeters = dict()
        for file_name, df_name in zip(file_names, get_dataframe_names(file_names, post_state.df_names)):

            delimeter = guess_delimeter(file_name)
            file_delimeters[df_name] = delimeter

            if delimeter != ',':
                log('used_non_standard_delimeter', {'delimeter': delimeter})

            df = pd.read_csv(file_name, sep=delimeter)
            renames = get_header_renames(df.keys())
            if len(renames) > 0:
                # Save that we did these renames
                column_header_renames[df_name] = renames
                # Actually perform any renames we need to
                df.rename(columns=renames, inplace=True)

            post_state.add_df_to_state(df, DATAFRAME_SOURCE_IMPORTED, df_name=df_name)        

        # Save the renames that have occured in the step, for transpilation reasons
        # and also save the seperator that we used for each file
        return post_state, {
            'column_header_renames': column_header_renames,
            'file_delimeters': file_delimeters
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        file_names: List[str]
    ) -> List[str]:
        code = ['import pandas as pd']
        for file_name, df_name in zip(file_names, post_state.df_names[len(post_state.df_names) - len(file_names):]):

            delimeter = execution_data['file_delimeters'][df_name]

            if delimeter == ',':
                # We don't add the delimeter if it's the default comma
                # NOTE: we add a r in front of the string so that it is a raw string
                # and file slashes are not interpreted as a unicode sequence
                code.append(
                    f'{df_name} = pd.read_csv(r\'{file_name}\')'
                )
            else:
                # If there is a delimeter for this file, we use it
                # NOTE: we add a r in front of the string so that it is a raw string
                # and file slashes are not interpreted as a unicode sequence
                code.append(
                    f'{df_name} = pd.read_csv(r\'{file_name}\', sep=\'{delimeter}\')'
                )
            # If we had to rename columns, mark that as well
            if df_name in execution_data['column_header_renames']:
                renames = execution_data['column_header_renames'][df_name]
                code.append(
                    f'{df_name}.rename(columns={json.dumps(renames)}, inplace=True)'
                )

        return code

    @classmethod
    def describe(
        cls,
        file_names: List[str],
        df_names=None,
        **params
    ) -> str:
        return f'Imported {", ".join(file_names)}'


def guess_delimeter(file_name: str):
    """
    Given a path to a file that is assumed to exist and be a CSV, this
    function guesses the delimeter that is used by that file
    """
    s = csv.Sniffer()
    with open(file_name, 'r') as f:
        return s.sniff(f.readline()).delimiter


def file_name_to_df_name(file_name: str):
    # First, we strip off any path beyond the final filename
    file_name = basename(normpath(file_name))

    # We abuse the fact that all valid mito headers are almost valid variable names
    # If they start w/ numbers, we add letters
    possible_var_name = make_valid_header(file_name)
    if len(possible_var_name) == 0 or possible_var_name[0].isnumeric():
        return 'df_' + possible_var_name
    return possible_var_name


def get_dataframe_names(file_names, existing_df_names):
    """
    Helper function for taking a list of file names and turning them into valid
    names for dataframes.

    NOTE:
    1. If there are duplicates, appends onto the end of them to deduplicate.
    2. Avoids overwriting existing df names.
    """
    new_names_inital = [file_name_to_df_name(file_name) for file_name in file_names]

    final_names = []

    # Keep appending names till we get one that doesn't overlap
    for name in new_names_inital:
        curr_name = name
        count = 0
        while curr_name in final_names or curr_name in existing_df_names:
            curr_name = f'{name}_{count}'
            count += 1
        
        final_names.append(curr_name)
    
    return final_names


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import deepcopy
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.state import State
from typing import Any, Dict, List, Optional
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
)

def get_valid_index(dfs, sheet_index, new_column_index):
    # make sure new_column_index is valid
    if new_column_index < 0:
        new_column_index = 0

    if new_column_index >= len(dfs[sheet_index].columns):
        new_column_index = len(dfs[sheet_index].columns) - 1

    return new_column_index


class ReorderColumnStepPerformer(StepPerformer):
    """""
    A reorder_column step, which allows you to move 
    a column to a different location in the df.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'reorder_column'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Reordered Columns'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'reorder_column_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        column_header: str,
        new_column_index: int,
        **params
    ) -> State:
        # if the sheet doesn't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error(sheet_index)

        # We check that the column to be reordered exists
        missing_column = set([column_header]).difference(prev_state.column_metatype[sheet_index].keys())
        if len(missing_column) > 0:
            raise make_no_column_error(missing_column)

        new_column_index = get_valid_index(prev_state.dfs, sheet_index, new_column_index)
            
        # Create a new post state
        post_state = deepcopy(prev_state)

        # Actually execute the column reordering
        post_state.dfs[sheet_index] = _execute_reorder_column(
            prev_state.dfs[sheet_index],
            column_header,
            new_column_index
        )

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        column_header: str,
        new_column_index: int
    ) -> List[str]:
        new_column_index = get_valid_index(prev_state.dfs, sheet_index, new_column_index)
        df_name = post_state.df_names[sheet_index]

        # get columns in df
        columns_list_line = f'{df_name}_columns = [col for col in {df_name}.columns if col != \'{column_header}\']'

        # insert column into correct location 
        insert_line = f'{df_name}_columns.insert({new_column_index}, \'{column_header}\')'
        
        # Apply reorder line
        apply_reorder_line = f'{df_name} = {df_name}[{df_name}_columns]'

        return [columns_list_line, insert_line, apply_reorder_line]

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        column_header: str,
        new_column_index: int,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_name = df_names[sheet_index]
            return f'Reordered {column_header} in {df_name}'
        return f'Reordered {column_header}'


def _execute_reorder_column(df, column_header, new_column_index):
    """
    Helper function for reordering a column in the dataframe
    """
    df_columns = [col for col in df.columns if col != column_header]
    df_columns.insert(new_column_index, column_header)
    return df[df_columns]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from mitosheet.sheet_functions.types.utils import NUMBER_SERIES, STRING_SERIES, get_mito_type
import pandas as pd
import pytest
import json

from mitosheet.tests.test_utils import create_mito_wrapper

COLUMN_FILTER_TYPE_TESTS = [
    ('int_column', NUMBER_SERIES),
    ('float_column', NUMBER_SERIES),
    ('mixed_float_and_int_column', NUMBER_SERIES),
    ('string_column', STRING_SERIES),
    ('mixed_string_and_int_column', STRING_SERIES),
]

@pytest.mark.parametrize('column_header, filter_type', COLUMN_FILTER_TYPE_TESTS)
def test_get_mito_type(column_header, filter_type):
    df = pd.DataFrame(data={
        'int_column': [1, 2, 3, 4, 5, 6],
        'float_column': [1.1, 2.2, 3.0, 4.5, 5.7, 6.9],
        'mixed_float_and_int_column': [1.1, 2, 3, 4, 5.7, 6.9],
        'string_column': ["1", "2", "3", "4", "5", "6"],
        'mixed_string_and_int_column': [1, 2, "3", 4, "5", "6"]
    })

    assert filter_type == get_mito_type(df[column_header])


def test_get_mito_type_after_formula():
    mito = create_mito_wrapper(['123'])
    mito.set_formula('=A', 0, 'B', add_column=True)
    assert STRING_SERIES == get_mito_type(mito.get_column(0, 'B', False))
    mito.set_formula('=100', 0, 'C', add_column=True)
    assert NUMBER_SERIES == get_mito_type(mito.get_column(0, 'C', False))


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Deletes an existing saved analysis.
"""

from mitosheet.saved_analyses import delete_saved_analysis

DELETE_SAVED_ANALYSIS_UPDATE_EVENT = 'delete_analysis_update'
DELETE_SAVED_ANALYSIS_UPDATE_PARAMS = [
    'analysis_name',
]

def execute_delete_saved_analysis_update(
        steps_manager,
        analysis_name,
    ):
    """
    This function deletes the saved analysis from the user's .mito folder
    """ 
    delete_saved_analysis(analysis_name)
    

DELETE_SAVED_ANALYSIS_UPDATE = {
    'event_type': DELETE_SAVED_ANALYSIS_UPDATE_EVENT,
    'params': DELETE_SAVED_ANALYSIS_UPDATE_PARAMS,
    'execute': execute_delete_saved_analysis_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
for tutorial v3
"""

MITO_TUTORIAL_CHECKPOINT_5 = {
    'name': 'checkpoint5',
    'saved_analysis': {
        "version": "0.1.96",
        "steps": {
            "1": {
                "step_version": 1,
                "step_type": "simple_import",
                "file_names": [
                    "AMTRAK-Stations-2010.csv",
                    "Zipcode-Data-2010.csv"
                ]
            },
            "2": {
                "step_version": 1,
                "step_type": "merge",
                "sheet_index_one": 0,
                "merge_key_one": "Zip",
                "selected_columns_one": [
                    "State",
                    "Checked_Bags",
                    "Zip"
                ],
                "sheet_index_two": 1,
                "merge_key_two": "Zip",
                "selected_columns_two": [
                    "Median_Income",
                    "Zip"
                ]
            },
            "3": {
                "step_version": 1,
                "step_type": "dataframe_rename",
                "sheet_index": 2,
                "old_dataframe_name": "df3",
                "new_dataframe_name": "Data"
            },
            "4": {
                "step_version": 1,
                "step_type": "add_column",
                "sheet_index": 2,
                "column_header": "E"
            },
            "5": {
                "step_version": 1,
                "step_type": "rename_column",
                "sheet_index": 2,
                "old_column_header": "E",
                "new_column_header": "Accepts_Bags"
            },
            "6": {
                "step_version": 1,
                "step_type": "set_column_formula",
                "sheet_index": 2,
                "column_header": "Accepts_Bags",
                "new_formula": "=IF(Checked_Bags=='Y', 1, 0)",
                "old_formula": "=0"
            }
        }
    }
}






#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.


from copy import copy
from typing import Any, Dict, List, Optional
from mitosheet.state import DATAFRAME_SOURCE_MERGED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
)


class MergeStepPerformer(StepPerformer):
    """
    Allows you to merge two dataframes together.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'merge'
    
    @classmethod
    def step_display_name(cls) -> str:
        return 'Merged Dataframes'

    @classmethod
    def step_event_type(cls) -> str:
        return 'merge_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        how: str,
        sheet_index_one: int,
        merge_key_one: str,
        selected_columns_one: List[str],
        sheet_index_two: int,
        merge_key_two: str,
        selected_columns_two: List[str],
        **params
    ) -> State:
        # if the sheets don't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index_one):
            raise make_no_sheet_error(sheet_index_one)

        if not prev_state.does_sheet_index_exist_within_state(sheet_index_two):
            raise make_no_sheet_error(sheet_index_two)

        # We check that the merge doesn't use any columns that don't exist
        missing_sheet_one_key = {merge_key_one}.difference(prev_state.column_metatype[sheet_index_one].keys())
        if any(missing_sheet_one_key):
            raise make_no_column_error(missing_sheet_one_key)

        missing_sheet_two_key = {merge_key_two}.difference(prev_state.column_metatype[sheet_index_two].keys())
        if any(missing_sheet_two_key):
            raise make_no_column_error(missing_sheet_two_key)

        # We create a shallow copy to make the new post state
        post_state = copy(prev_state)

        new_df = _execute_merge(
            prev_state.dfs,
            prev_state.df_names,
            how,
            sheet_index_one,
            merge_key_one,
            selected_columns_one,
            sheet_index_two,
            merge_key_two,
            selected_columns_two
        )    

        # Add this dataframe to the new post state
        post_state.add_df_to_state(new_df, DATAFRAME_SOURCE_MERGED)

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        how: str,
        sheet_index_one: int,
        merge_key_one: str,
        selected_columns_one: List[str],
        sheet_index_two: int,
        merge_key_two: str,
        selected_columns_two: List[str]
    ) -> List[str]:

        # update df indexes to start at 1
        df_one_name = post_state.df_names[sheet_index_one]
        df_two_name = post_state.df_names[sheet_index_two]
        df_new_name = post_state.df_names[len(post_state.dfs) - 1]

        # Now, we build the merge code 
        merge_code = []
        if how == 'lookup':
            # If the mege is a lookup, then we add the drop duplicates code
            temp_df_name = 'temp_df'
            merge_code.append(f'{temp_df_name} = {df_two_name}.drop_duplicates(subset=\'{merge_key_two}\') # Remove duplicates so lookup merge only returns first match')
            how_to_use = 'left'
        else:
            temp_df_name = df_two_name
            how_to_use = how


        # If we are only taking some columns, write the code to drop the ones we don't need!
        deleted_columns_one = set(post_state.dfs[sheet_index_one].keys()).difference(set(selected_columns_one))
        deleted_columns_two = set(post_state.dfs[sheet_index_two].keys()).difference(set(selected_columns_two))
        if len(deleted_columns_one) > 0:
            merge_code.append(
                f'{df_one_name}_tmp = {df_one_name}.drop({list(deleted_columns_one)}, axis=1)'
            )
        if len(deleted_columns_two) > 0:
            merge_code.append(
                f'{df_two_name}_tmp = {temp_df_name}.drop({list(deleted_columns_two)}, axis=1)'
            )

        # If we drop columns, we merge the new dataframes
        df_one_to_merge = df_one_name if len(deleted_columns_one) == 0 else f'{df_one_name}_tmp'
        df_two_to_merge = temp_df_name if len(deleted_columns_two) == 0 else f'{df_two_name}_tmp'

        # We insist column names are unique in dataframes, so while we default the suffixes to be the dataframe
        # names
        suffix_one = df_one_name
        suffix_two = df_two_name if df_two_name != df_one_name else f'{df_two_name}_2'

        # Finially append the merge
        merge_code.append(
            f'{df_new_name} = {df_one_to_merge}.merge({df_two_to_merge}, left_on=[\'{merge_key_one}\'], right_on=[\'{merge_key_two}\'], how=\'{how_to_use}\', suffixes=[\'_{suffix_one}\', \'_{suffix_two}\'])'
        )

        # And then return it
        return merge_code

    @classmethod
    def describe(
        cls,
        how: str,
        sheet_index_one: int,
        merge_key_one: str,
        selected_columns_one: List[str],
        sheet_index_two: int,
        merge_key_two: str,
        selected_columns_two: List[str],
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_one_name = df_names[sheet_index_one]
            df_two_name = df_names[sheet_index_two]
            return f'Merged {df_one_name} and {df_two_name}'
        return f'Merged dataframes {sheet_index_one} and {sheet_index_two}'

def _execute_merge(
        dfs, 
        df_names,
        how,
        sheet_index_one,
        merge_key_one, 
        selected_columns_one,
        sheet_index_two,
        merge_key_two,
        selected_columns_two
    ):
    """
    Executes a merge on the sheets with the given indexes, merging on the 
    given keys, and only keeping the selection columns from each df.
    """
    if how == 'lookup':
        # We drop duplicates to avoid pairwise duplication on the merge.
        temp_df = dfs[sheet_index_two].drop_duplicates(subset=merge_key_two)
        # We overwrite the how variable to 'left' so it can be used in the merge
        how_to_use = 'left'
    else:
        temp_df = dfs[sheet_index_two]
        how_to_use = how

    # Then we delete all the columns from each we don't wanna keep
    deleted_columns_one = set(dfs[sheet_index_one].keys()).difference(set(selected_columns_one))
    deleted_columns_two = set(dfs[sheet_index_two].keys()).difference(set(selected_columns_two))

    df_one_cleaned = dfs[sheet_index_one].drop(deleted_columns_one, axis=1)
    df_two_cleaned = temp_df.drop(deleted_columns_two, axis=1)

    # Finially, we perform the merge!
    df_one_name = df_names[sheet_index_one]
    df_two_name = df_names[sheet_index_two]
    # We make sure the suffixes aren't the same, as otherwise we might end up with 
    # one df with duplicated column headers
    suffix_one = df_one_name
    suffix_two = df_two_name if df_two_name != df_one_name else f'{df_two_name}_2'

    return df_one_cleaned.merge(df_two_cleaned, left_on=[merge_key_one], right_on=[merge_key_two], how=how_to_use, suffixes=[f'_{suffix_one}', f'_{suffix_two}'])